@@ -35,6 +35,9 @@
 
 #include "gdk-pixbuf-private.h"
 #include "gdk-pixbuf-io.h"
+#include "gdk-pixbuf-alias.h"
+
+#include <glib/gstdio.h>
 
 #ifdef G_OS_WIN32
 #define STRICT
@@ -75,15 +78,41 @@ format_check (GdkPixbufModule *module, guchar *buffer, int size)
 	return 0;
 }
 
+G_LOCK_DEFINE_STATIC (init_lock);
+G_LOCK_DEFINE_STATIC (threadunsafe_loader_lock);
+
+gboolean
+_gdk_pixbuf_lock (GdkPixbufModule *image_module)
+{
+ 	if (g_threads_got_initialized &&
+	    !(image_module->info->flags & GDK_PIXBUF_FORMAT_THREADSAFE)) {
+ 		G_LOCK (threadunsafe_loader_lock);
+
+		return TRUE;
+ 	}
+
+	return FALSE;
+}
+ 
+void
+_gdk_pixbuf_unlock (GdkPixbufModule *image_module)
+{
+	if (!(image_module->info->flags & GDK_PIXBUF_FORMAT_THREADSAFE)) {
+		G_UNLOCK (threadunsafe_loader_lock);
+	}
+}
+
 static GSList *file_formats = NULL;
 
 static void gdk_pixbuf_io_init ();
 
 static GSList *
-get_file_formats ()
+get_file_formats (void)
 {
+	G_LOCK (init_lock);
 	if (file_formats == NULL)
 		gdk_pixbuf_io_init ();
+	G_UNLOCK (init_lock);
 	
 	return file_formats;
 }
@@ -221,7 +250,7 @@ correct_prefix (gchar **path)
     }
 }
 
-#endif
+#endif  /* G_OS_WIN32 */
 
 static gchar *
 gdk_pixbuf_get_module_file (void)
@@ -235,7 +264,7 @@ gdk_pixbuf_get_module_file (void)
 }
 
 static void 
-gdk_pixbuf_io_init ()
+gdk_pixbuf_io_init (void)
 {
 	GIOChannel *channel;
 	gchar *line_buf;
@@ -333,7 +362,7 @@ gdk_pixbuf_io_init ()
 					module->info->mime_types =
 						g_realloc (module->info->mime_types, (n + 1) * sizeof (gchar*));
 					module->info->mime_types[n - 1] = g_strdup (tmp_buf->str);
-					module->info->mime_types[n] = 0;
+					module->info->mime_types[n] = NULL;
 					n++;
 				}
 			}
@@ -346,7 +375,7 @@ gdk_pixbuf_io_init ()
 					module->info->extensions =
 						g_realloc (module->info->extensions, (n + 1) * sizeof (gchar*));
 					module->info->extensions[n - 1] = g_strdup (tmp_buf->str);
-					module->info->extensions[n] = 0;
+					module->info->extensions[n] = NULL;
 					n++;
 				}
 			}
@@ -395,14 +424,14 @@ gdk_pixbuf_io_init ()
 /* actually load the image handler - gdk_pixbuf_get_module only get a */
 /* reference to the module to load, it doesn't actually load it       */
 /* perhaps these actions should be combined in one function           */
-gboolean
-_gdk_pixbuf_load_module (GdkPixbufModule *image_module,
-                         GError         **error)
+static gboolean
+_gdk_pixbuf_load_module_unlocked (GdkPixbufModule *image_module,
+				  GError         **error)
 {
 	char *path;
 	GModule *module;
 	gpointer sym;
-	
+		
         g_return_val_if_fail (image_module->module == NULL, FALSE);
 
 	path = image_module->module_path;
@@ -432,7 +461,29 @@ _gdk_pixbuf_load_module (GdkPixbufModule *image_module,
                 return FALSE;
         }
 }
-#else
+
+gboolean
+_gdk_pixbuf_load_module (GdkPixbufModule *image_module,
+			 GError         **error)
+{
+	gboolean ret;
+	gboolean locked = FALSE;
+
+	/* be extra careful, maybe the module initializes
+	 * the thread system
+	 */
+	if (g_threads_got_initialized)
+	{
+		G_LOCK (init_lock);
+		locked = TRUE;
+	}
+	ret = _gdk_pixbuf_load_module_unlocked (image_module, error);
+	if (locked)
+		G_UNLOCK (init_lock);
+	return ret;
+}
+
+#else  /* !USE_GMODULE */
 
 #define module(type) \
   extern void MODULE_ENTRY (type, fill_info)   (GdkPixbufFormat *info);   \
@@ -592,7 +643,7 @@ gdk_pixbuf_io_init ()
 	};
 	gchar **name;
 	GdkPixbufModule *module = NULL;
-	
+
 	for (name = included_formats; *name; name++) {
 		module = g_new0 (GdkPixbufModule, 1);
 		module->module_name = *name;
@@ -603,7 +654,7 @@ gdk_pixbuf_io_init ()
 	}
 }
 
-#endif
+#endif  /* !USE_GMODULE */
 
 
 
@@ -641,7 +692,7 @@ _gdk_pixbuf_get_module (guchar *buffer, guint size,
 
 	gint score, best = 0;
 	GdkPixbufModule *selected = NULL;
-	gchar *utf8_filename = NULL;
+	gchar *display_name = NULL;
 
 	for (modules = get_file_formats (); modules; modules = g_slist_next (modules)) {
 		GdkPixbufModule *module = (GdkPixbufModule *)modules->data;
@@ -661,16 +712,14 @@ _gdk_pixbuf_get_module (guchar *buffer, guint size,
 		return selected;
 
         if (filename)
-		utf8_filename = g_filename_to_utf8 (filename, -1,
-						    NULL, NULL, NULL);
-	
-	if (utf8_filename) {
+	{
+		display_name = g_filename_display_name (filename);
 		g_set_error (error,
 			     GDK_PIXBUF_ERROR,
 			     GDK_PIXBUF_ERROR_UNKNOWN_TYPE,
 			     _("Couldn't recognize the image file format for file '%s'"),
-			     utf8_filename);
-		g_free (utf8_filename);
+			     display_name);
+		g_free (display_name);
 	}
         else
                 g_set_error (error,
@@ -703,56 +752,57 @@ _gdk_pixbuf_generic_image_load (GdkPixbufModule *module,
 	GdkPixbuf *pixbuf = NULL;
 	GdkPixbufAnimation *animation = NULL;
 	gpointer context;
+	gboolean locked;
 
-	if (module->load != NULL)
-		return (* module->load) (f, error);
-	
-	if (module->begin_load != NULL) {
+	locked = _gdk_pixbuf_lock (module);
+
+	if (module->load != NULL) {
+		pixbuf = (* module->load) (f, error);
+	} else if (module->begin_load != NULL) {
 		
 		context = module->begin_load (NULL, prepared_notify, NULL, &pixbuf, error);
 	
 		if (!context)
-			return NULL;
+			goto out;
 		
 		while (!feof (f) && !ferror (f)) {
 			length = fread (buffer, 1, sizeof (buffer), f);
 			if (length > 0)
 				if (!module->load_increment (context, buffer, length, error)) {
 					module->stop_load (context, NULL);
-					if (pixbuf != NULL)
+					if (pixbuf != NULL) {
 						g_object_unref (pixbuf);
-					return NULL;
+						pixbuf = NULL;
+					}
+					goto out;
 				}
 		}
 		
 		if (!module->stop_load (context, error)) {
-			if (pixbuf != NULL)
+			if (pixbuf != NULL) {
 				g_object_unref (pixbuf);
-			return NULL;
+				pixbuf = NULL;
+			}
 		}
-		
-		return pixbuf;
-	}
-	
-	if (module->load_animation != NULL) {
+	} else if (module->load_animation != NULL) {
 		animation = (* module->load_animation) (f, error);
 		if (animation != NULL) {
 			pixbuf = gdk_pixbuf_animation_get_static_image (animation);
 
 			g_object_ref (pixbuf);
-
 			g_object_unref (animation);
-			
-			return pixbuf;
 		}
 	}
 
-	return NULL;
+ out:
+	if (locked)
+		_gdk_pixbuf_unlock (module);
+	return pixbuf;
 }
 
 /**
  * gdk_pixbuf_new_from_file:
- * @filename: Name of file to load.
+ * @filename: Name of file to load, in the GLib file name encoding
  * @error: Return location for an error
  *
  * Creates a new pixbuf by loading an image from a file.  The file format is
@@ -773,46 +823,47 @@ gdk_pixbuf_new_from_file (const char *filename,
 	FILE *f;
 	guchar buffer [128];
 	GdkPixbufModule *image_module;
+	gchar *display_name;
 
 	g_return_val_if_fail (filename != NULL, NULL);
         g_return_val_if_fail (error == NULL || *error == NULL, NULL);
+	
+	display_name = g_filename_display_name (filename);	
 
-	f = fopen (filename, "rb");
+	f = g_fopen (filename, "rb");
 	if (!f) {
-                gchar *utf8_filename = g_filename_to_utf8 (filename, -1,
-                                                           NULL, NULL, NULL);
                 g_set_error (error,
                              G_FILE_ERROR,
                              g_file_error_from_errno (errno),
                              _("Failed to open file '%s': %s"),
-                             utf8_filename ? utf8_filename : "???",
+                             display_name,
                              g_strerror (errno));
-                g_free (utf8_filename);
+                g_free (display_name);
 		return NULL;
         }
 
 	size = fread (&buffer, 1, sizeof (buffer), f);
 	if (size == 0) {
-                gchar *utf8_filename = g_filename_to_utf8 (filename, -1,
-                                                           NULL, NULL, NULL);
                 g_set_error (error,
                              GDK_PIXBUF_ERROR,
                              GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
                              _("Image file '%s' contains no data"),
-                             utf8_filename ? utf8_filename : "???");
-                g_free (utf8_filename);
+                             display_name);
+                g_free (display_name);
 		fclose (f);
 		return NULL;
 	}
 
 	image_module = _gdk_pixbuf_get_module (buffer, size, filename, error);
         if (image_module == NULL) {
+                g_free (display_name);
                 fclose (f);
                 return NULL;
         }
 
 	if (image_module->module == NULL)
                 if (!_gdk_pixbuf_load_module (image_module, error)) {
+			g_free (display_name);
                         fclose (f);
                         return NULL;
                 }
@@ -829,37 +880,51 @@ gdk_pixbuf_new_from_file (const char *filename,
                  * the invariant that error gets set if NULL is returned.
                  */
 
-                gchar *utf8_filename = g_filename_to_utf8 (filename, -1,
-                                                           NULL, NULL, NULL);
-
                 g_warning ("Bug! gdk-pixbuf loader '%s' didn't set an error on failure.", image_module->module_name);
                 g_set_error (error,
                              GDK_PIXBUF_ERROR,
                              GDK_PIXBUF_ERROR_FAILED,
                              _("Failed to load image '%s': reason not known, probably a corrupt image file"),
-                             utf8_filename ? utf8_filename : "???");
-                g_free (utf8_filename);
+                             display_name);
         } else if (error != NULL && *error != NULL) {
 
           /* Add the filename to the error message */
           GError *e = *error;
           gchar  *old;
-          gchar  *utf8_filename = g_filename_to_utf8 (filename, -1,
-                                                      NULL, NULL, NULL);
 
           old = e->message;
-
           e->message = g_strdup_printf (_("Failed to load image '%s': %s"),
-                                        utf8_filename ? utf8_filename : "???",
+                                        display_name,
                                         old);
-
-          g_free (utf8_filename);
           g_free (old);
         }
 
+	g_free (display_name);
 	return pixbuf;
 }
 
+#ifdef G_OS_WIN32
+
+#undef gdk_pixbuf_new_from_file
+GdkPixbuf *
+gdk_pixbuf_new_from_file (const char *filename,
+                          GError    **error)
+{
+	gchar *utf8_filename =
+		g_locale_to_utf8 (filename, -1, NULL, NULL, error);
+	GdkPixbuf *retval;
+
+	if (utf8_filename == NULL)
+		return NULL;
+
+	retval = gdk_pixbuf_new_from_file_utf8 (utf8_filename, error);
+
+	g_free (utf8_filename);
+
+	return retval;
+}
+#endif
+
 static void
 size_prepared_cb (GdkPixbufLoader *loader, 
 		  int              width,
@@ -867,27 +932,33 @@ size_prepared_cb (GdkPixbufLoader *loader,
 		  gpointer         data)
 {
 	struct {
-		int width;
-		int height;
+		gint width;
+		gint height;
+		gboolean preserve_aspect_ratio;
 	} *info = data;
 
 	g_return_if_fail (width > 0 && height > 0);
 
-	if ((double)height * (double)info->width >
-	    (double)width * (double)info->height) {
-		width = 0.5 + (double)width * (double)info->height / (double)height;
-		height = info->height;
+	if(info->preserve_aspect_ratio) {
+		if ((double)height * (double)info->width >
+		    (double)width * (double)info->height) {
+			width = 0.5 + (double)width * (double)info->height / (double)height;
+			height = info->height;
+		} else {
+			height = 0.5 + (double)height * (double)info->width / (double)width;
+			width = info->width;
+		}
 	} else {
-		height = 0.5 + (double)height * (double)info->width / (double)width;
 		width = info->width;
+		height = info->height;
 	}
-
+	
 	gdk_pixbuf_loader_set_size (loader, width, height);
 }
 
 /**
  * gdk_pixbuf_new_from_file_at_size:
- * @filename: Name of file to load.
+ * @filename: Name of file to load, in the GLib file name encoding
  * @width: The width the image should have
  * @height: The height the image should have
  * @error: Return location for an error
@@ -895,12 +966,14 @@ size_prepared_cb (GdkPixbufLoader *loader,
  * Creates a new pixbuf by loading an image from a file.  The file format is
  * detected automatically. If %NULL is returned, then @error will be set.
  * Possible errors are in the #GDK_PIXBUF_ERROR and #G_FILE_ERROR domains.
- * The image will be scaled to fit in the requested size, preserving its aspect ratio.
+ * The image will be scaled to fit in the requested size, preserving
+ * the image's aspect ratio.
  *
- * Return value: A newly-created pixbuf with a reference count of 1, or %NULL if
- * any of several error conditions occurred:  the file could not be opened,
- * there was no loader for the file's format, there was not enough memory to
- * allocate the image buffer, or the image file contained invalid data.
+ * Return value: A newly-created pixbuf with a reference count of 1, or 
+ * %NULL if any of several error conditions occurred:  the file could not 
+ * be opened, there was no loader for the file's format, there was not 
+ * enough memory to allocate the image buffer, or the image file contained 
+ * invalid data.
  *
  * Since: 2.4
  **/
@@ -910,6 +983,67 @@ gdk_pixbuf_new_from_file_at_size (const char *filename,
 				  int         height,
 				  GError    **error)
 {
+	return gdk_pixbuf_new_from_file_at_scale (filename, 
+						  width, height, 
+						  TRUE, error);
+}
+
+#ifdef G_OS_WIN32
+
+#undef gdk_pixbuf_new_from_file_at_size
+
+GdkPixbuf *
+gdk_pixbuf_new_from_file_at_size (const char *filename,
+				  int         width, 
+				  int         height,
+				  GError    **error)
+{
+	gchar *utf8_filename =
+		g_locale_to_utf8 (filename, -1, NULL, NULL, error);
+	GdkPixbuf *retval;
+
+	if (utf8_filename == NULL)
+		return NULL;
+
+	retval = gdk_pixbuf_new_from_file_at_size_utf8 (utf8_filename,
+							width, height,
+							error);
+
+	g_free (utf8_filename);
+
+	return retval;
+}
+#endif
+
+/**
+ * gdk_pixbuf_new_from_file_at_scale:
+ * @filename: Name of file to load, in the GLib file name encoding
+ * @width: The width the image should have
+ * @height: The height the image should have
+ * @preserve_aspect_ratio: %TRUE to preserve the image's aspect ratio
+ * @error: Return location for an error
+ *
+ * Creates a new pixbuf by loading an image from a file.  The file format is
+ * detected automatically. If %NULL is returned, then @error will be set.
+ * Possible errors are in the #GDK_PIXBUF_ERROR and #G_FILE_ERROR domains.
+ * The image will be scaled to fit in the requested size, optionally preserving
+ * the image's aspect ratio.
+ *
+ * Return value: A newly-created pixbuf with a reference count of 1, or %NULL 
+ * if any of several error conditions occurred:  the file could not be opened,
+ * there was no loader for the file's format, there was not enough memory to
+ * allocate the image buffer, or the image file contained invalid data.
+ *
+ * Since: 2.6
+ **/
+GdkPixbuf *
+gdk_pixbuf_new_from_file_at_scale (const char *filename,
+				   int         width, 
+				   int         height,
+				   gboolean    preserve_aspect_ratio,
+				   GError    **error)
+{
+
 	GdkPixbufLoader *loader;
 	GdkPixbuf       *pixbuf;
 
@@ -919,22 +1053,22 @@ gdk_pixbuf_new_from_file_at_size (const char *filename,
 	struct {
 		gint width;
 		gint height;
+		gboolean preserve_aspect_ratio;
 	} info;
 
 	g_return_val_if_fail (filename != NULL, NULL);
         g_return_val_if_fail (width > 0 && height > 0, NULL);
 
-	f = fopen (filename, "rb");
+	f = g_fopen (filename, "rb");
 	if (!f) {
-                gchar *utf8_filename = g_filename_to_utf8 (filename, -1,
-                                                           NULL, NULL, NULL);
+                gchar *display_name = g_filename_display_name (filename);
                 g_set_error (error,
                              G_FILE_ERROR,
                              g_file_error_from_errno (errno),
                              _("Failed to open file '%s': %s"),
-                             utf8_filename ? utf8_filename : "???",
+                             display_name,
                              g_strerror (errno));
-                g_free (utf8_filename);
+                g_free (display_name);
 		return NULL;
         }
 
@@ -942,6 +1076,7 @@ gdk_pixbuf_new_from_file_at_size (const char *filename,
 
 	info.width = width;
 	info.height = height;
+        info.preserve_aspect_ratio = preserve_aspect_ratio;
 
 	g_signal_connect (loader, "size-prepared", G_CALLBACK (size_prepared_cb), &info);
 
@@ -966,17 +1101,14 @@ gdk_pixbuf_new_from_file_at_size (const char *filename,
 	pixbuf = gdk_pixbuf_loader_get_pixbuf (loader);
 
 	if (!pixbuf) {
-                gchar *utf8_filename = g_filename_to_utf8 (filename, -1,
-                                                           NULL, NULL, NULL);
-
+                gchar *display_name = g_filename_display_name (filename);
 		g_object_unref (loader);
-
                 g_set_error (error,
                              GDK_PIXBUF_ERROR,
                              GDK_PIXBUF_ERROR_FAILED,
                              _("Failed to load image '%s': reason not known, probably a corrupt image file"),
-                             utf8_filename ? utf8_filename : "???");
-                g_free (utf8_filename);
+                             display_name);
+                g_free (display_name);
 		return NULL;
 	}
 
@@ -987,6 +1119,36 @@ gdk_pixbuf_new_from_file_at_size (const char *filename,
 	return pixbuf;
 }
 
+#ifdef G_OS_WIN32
+
+#undef gdk_pixbuf_new_from_file_at_scale
+
+GdkPixbuf *
+gdk_pixbuf_new_from_file_at_scale (const char *filename,
+				   int         width, 
+				   int         height,
+				   gboolean    preserve_aspect_ratio,
+				   GError    **error)
+{
+	gchar *utf8_filename =
+		g_locale_to_utf8 (filename, -1, NULL, NULL, error);
+	GdkPixbuf *retval;
+
+	if (utf8_filename == NULL)
+		return NULL;
+
+	retval = gdk_pixbuf_new_from_file_at_scale_utf8 (utf8_filename,
+							 width, height,
+							 preserve_aspect_ratio,
+							 error);
+
+	g_free (utf8_filename);
+
+	return retval;
+}
+#endif
+
+
 static void
 info_cb (GdkPixbufLoader *loader, 
 	 int              width,
@@ -1039,7 +1201,7 @@ gdk_pixbuf_get_file_info (const gchar  *filename,
 
 	g_return_val_if_fail (filename != NULL, NULL);
 
-	f = fopen (filename, "rb");
+	f = g_fopen (filename, "rb");
 	if (!f)
 		return NULL;
 
@@ -1088,7 +1250,10 @@ gdk_pixbuf_new_from_xpm_data (const char **data)
 	GdkPixbuf *(* load_xpm_data) (const char **data);
 	GdkPixbuf *pixbuf;
         GError *error = NULL;
-	GdkPixbufModule *xpm_module = _gdk_pixbuf_get_named_module ("xpm", &error);
+	GdkPixbufModule *xpm_module;
+	gboolean locked;
+
+	xpm_module = _gdk_pixbuf_get_named_module ("xpm", &error);
 	if (xpm_module == NULL) {
 		g_warning ("Error loading XPM image loader: %s", error->message);
 		g_error_free (error);
@@ -1102,14 +1267,19 @@ gdk_pixbuf_new_from_xpm_data (const char **data)
                         return NULL;
                 }
         }
-          
+
+	locked = _gdk_pixbuf_lock (xpm_module);
+
 	if (xpm_module->load_xpm_data == NULL) {
 		g_warning ("gdk-pixbuf XPM module lacks XPM data capability");
-		return NULL;
-	} else
+		pixbuf = NULL;
+	} else {
 		load_xpm_data = xpm_module->load_xpm_data;
-
-	pixbuf = (* load_xpm_data) (data);
+		pixbuf = (* load_xpm_data) (data);
+	}
+	
+	if (locked)
+		_gdk_pixbuf_unlock (xpm_module);
 	return pixbuf;
 }
 
@@ -1178,39 +1348,45 @@ gdk_pixbuf_real_save (GdkPixbuf     *pixbuf,
                       gchar        **values,
                       GError       **error)
 {
-       GdkPixbufModule *image_module = NULL;       
+	gboolean ret;
+	GdkPixbufModule *image_module = NULL;       
+	gboolean locked;
 
-       image_module = _gdk_pixbuf_get_named_module (type, error);
+	image_module = _gdk_pixbuf_get_named_module (type, error);
 
-       if (image_module == NULL)
-               return FALSE;
+	if (image_module == NULL)
+		return FALSE;
        
-       if (image_module->module == NULL)
-               if (!_gdk_pixbuf_load_module (image_module, error))
-                       return FALSE;
+	if (image_module->module == NULL)
+		if (!_gdk_pixbuf_load_module (image_module, error))
+			return FALSE;
 
-       if (image_module->save) {
-	       /* save normally */
-	       return (* image_module->save) (filehandle, pixbuf,
+	locked = _gdk_pixbuf_lock (image_module);
+
+	if (image_module->save) {
+		/* save normally */
+		ret = (* image_module->save) (filehandle, pixbuf,
 					      keys, values,
 					      error);
-       }
-       else if (image_module->save_to_callback) {
-	       /* save with simple callback */
-	       return (* image_module->save_to_callback) (save_to_file_callback,
+	} else if (image_module->save_to_callback) {
+		/* save with simple callback */
+		ret = (* image_module->save_to_callback) (save_to_file_callback,
 							  filehandle, pixbuf,
 							  keys, values,
 							  error);
-       }
-       else {
-	       /* can't save */
-               g_set_error (error,
-                            GDK_PIXBUF_ERROR,
-                            GDK_PIXBUF_ERROR_UNSUPPORTED_OPERATION,
-                            _("This build of gdk-pixbuf does not support saving the image format: %s"),
-                            type);
-               return FALSE;
-       }
+	} else {
+		/* can't save */
+		g_set_error (error,
+			     GDK_PIXBUF_ERROR,
+			     GDK_PIXBUF_ERROR_UNSUPPORTED_OPERATION,
+			     _("This build of gdk-pixbuf does not support saving the image format: %s"),
+			     type);
+		ret = FALSE;
+	}
+
+	if (locked)
+		_gdk_pixbuf_unlock (image_module);
+	return ret;
 }
 
 #define TMP_FILE_BUF_SIZE 4096
@@ -1230,6 +1406,7 @@ save_to_callback_with_tmp_file (GdkPixbufModule   *image_module,
 	gchar *buf = NULL;
 	gsize n;
 	gchar *filename = NULL;
+	gboolean locked;
 
 	buf = g_try_malloc (TMP_FILE_BUF_SIZE);
 	if (buf == NULL) {
@@ -1251,8 +1428,14 @@ save_to_callback_with_tmp_file (GdkPixbufModule   *image_module,
 			     _("Failed to open temporary file"));
 		goto end;
 	}
-	if (!(* image_module->save) (f, pixbuf, keys, values, error))
+
+	locked = _gdk_pixbuf_lock (image_module);
+	retval = (image_module->save) (f, pixbuf, keys, values, error);
+	if (locked)
+		_gdk_pixbuf_unlock (image_module);
+	if (!retval)
 		goto end;
+
 	rewind (f);
 	for (;;) {
 		n = fread (buf, 1, TMP_FILE_BUF_SIZE, f);
@@ -1277,7 +1460,7 @@ save_to_callback_with_tmp_file (GdkPixbufModule   *image_module,
 	if (f)
 		fclose (f);
 	if (filename) {
-		unlink (filename);
+		g_unlink (filename);
 		g_free (filename);
 	}
 	g_free (buf);
@@ -1294,39 +1477,45 @@ gdk_pixbuf_real_save_to_callback (GdkPixbuf         *pixbuf,
 				  gchar            **values,
 				  GError           **error)
 {
-       GdkPixbufModule *image_module = NULL;       
+	gboolean ret;
+	GdkPixbufModule *image_module = NULL;       
+	gboolean locked;
 
-       image_module = _gdk_pixbuf_get_named_module (type, error);
+	image_module = _gdk_pixbuf_get_named_module (type, error);
 
-       if (image_module == NULL)
-               return FALSE;
+	if (image_module == NULL)
+		return FALSE;
        
-       if (image_module->module == NULL)
-               if (!_gdk_pixbuf_load_module (image_module, error))
-                       return FALSE;
+	if (image_module->module == NULL)
+		if (!_gdk_pixbuf_load_module (image_module, error))
+			return FALSE;
 
-       if (image_module->save_to_callback) {
-	       /* save normally */
-	       return (* image_module->save_to_callback) (save_func, user_data, 
+	locked = _gdk_pixbuf_lock (image_module);
+
+	if (image_module->save_to_callback) {
+		/* save normally */
+		ret = (* image_module->save_to_callback) (save_func, user_data, 
 							  pixbuf, keys, values,
 							  error);
-       }
-       else if (image_module->save) {
-	       /* use a temporary file */
-	       return save_to_callback_with_tmp_file (image_module, pixbuf,
+	} else if (image_module->save) {
+		/* use a temporary file */
+		ret = save_to_callback_with_tmp_file (image_module, pixbuf,
 						      save_func, user_data, 
 						      keys, values,
 						      error);
-       }
-       else {
-	       /* can't save */
-               g_set_error (error,
-                            GDK_PIXBUF_ERROR,
-                            GDK_PIXBUF_ERROR_UNSUPPORTED_OPERATION,
-                            _("This build of gdk-pixbuf does not support saving the image format: %s"),
-                            type);
-               return FALSE;
-       }
+	} else {
+		/* can't save */
+		g_set_error (error,
+			     GDK_PIXBUF_ERROR,
+			     GDK_PIXBUF_ERROR_UNSUPPORTED_OPERATION,
+			     _("This build of gdk-pixbuf does not support saving the image format: %s"),
+			     type);
+		ret = FALSE;
+	}
+
+	if (locked)
+		_gdk_pixbuf_unlock (image_module);
+	return ret;
 }
 
  
@@ -1338,8 +1527,8 @@ gdk_pixbuf_real_save_to_callback (GdkPixbuf         *pixbuf,
  * @error: return location for error, or %NULL
  * @Varargs: list of key-value save options
  *
- * Saves pixbuf to a file in format @type. By default, "jpeg", "png" and 
- * "ico" are possible file formats to save in, but more formats may be
+ * Saves pixbuf to a file in format @type. By default, "jpeg", "png", "ico" 
+ * and "bmp" are possible file formats to save in, but more formats may be
  * installed. The list of all writable formats can be determined in the 
  * following way:
  *
@@ -1367,11 +1556,15 @@ gdk_pixbuf_real_save_to_callback (GdkPixbuf         *pixbuf,
  *                  "quality", "100", NULL);
  * </programlisting></informalexample>
  *
- * Currently only few parameters exist. JPEG images can be saved with a 
- * "quality" parameter; its value should be in the range [0,100]. 
+ * Currently only few parameters exist. JPEG images can be saved with a
+ * "quality" parameter; its value should be in the range [0,100].
+ *
  * Text chunks can be attached to PNG images by specifying parameters of
  * the form "tEXt::key", where key is an ASCII string of length 1-79.
- * The values are UTF-8 encoded strings. 
+ * The values are UTF-8 encoded strings. The PNG compression level can
+ * be specified using the "compression" parameter; it's value is in an
+ * integer in the range of [0,9].
+ *
  * ICO images can be saved in depth 16, 24, or 32, by using the "depth"
  * parameter. When the ICO saver is given "x_hot" and "y_hot" parameters,
  * it produces a CUR instead of an ICO.
@@ -1409,6 +1602,50 @@ gdk_pixbuf_save (GdkPixbuf  *pixbuf,
         return result;
 }
 
+#ifdef G_OS_WIN32
+
+#undef gdk_pixbuf_save
+
+gboolean
+gdk_pixbuf_save (GdkPixbuf  *pixbuf, 
+                 const char *filename, 
+                 const char *type, 
+                 GError    **error,
+                 ...)
+{
+	char *utf8_filename;
+        gchar **keys = NULL;
+        gchar **values = NULL;
+        va_list args;
+	gboolean result;
+
+        g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+        
+	utf8_filename = g_locale_to_utf8 (filename, -1, NULL, NULL, error);
+
+	if (utf8_filename == NULL)
+		return FALSE;
+
+        va_start (args, error);
+        
+        collect_save_options (args, &keys, &values);
+        
+        va_end (args);
+
+        result = gdk_pixbuf_savev_utf8 (pixbuf, utf8_filename, type,
+					keys, values,
+					error);
+
+	g_free (utf8_filename);
+
+        g_strfreev (keys);
+        g_strfreev (values);
+
+        return result;
+}
+
+#endif
+
 /**
  * gdk_pixbuf_savev:
  * @pixbuf: a #GdkPixbuf.
@@ -1418,7 +1655,7 @@ gdk_pixbuf_save (GdkPixbuf  *pixbuf,
  * @option_values: values for named options
  * @error: return location for error, or %NULL
  *
- * Saves pixbuf to a file in @type, which is currently "jpeg", "png" or "ico".
+ * Saves pixbuf to a file in @type, which is currently "jpeg", "png", "ico" or "bmp".
  * If @error is set, %FALSE will be returned. 
  * See gdk_pixbuf_save () for more details.
  *
@@ -1435,24 +1672,22 @@ gdk_pixbuf_savev (GdkPixbuf  *pixbuf,
 {
         FILE *f = NULL;
         gboolean result;
-        
        
         g_return_val_if_fail (filename != NULL, FALSE);
         g_return_val_if_fail (type != NULL, FALSE);
         g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
        
-        f = fopen (filename, "wb");
+        f = g_fopen (filename, "wb");
         
         if (f == NULL) {
-                gchar *utf8_filename = g_filename_to_utf8 (filename, -1,
-                                                           NULL, NULL, NULL);
+                gchar *display_name = g_filename_display_name (filename);
                 g_set_error (error,
                              G_FILE_ERROR,
                              g_file_error_from_errno (errno),
                              _("Failed to open '%s' for writing: %s"),
-                             utf8_filename ? utf8_filename : "???",
+                             display_name,
                              g_strerror (errno));
-                g_free (utf8_filename);
+                g_free (display_name);
                 return FALSE;
         }
 
@@ -1469,21 +1704,52 @@ gdk_pixbuf_savev (GdkPixbuf  *pixbuf,
        }
 
        if (fclose (f) < 0) {
-               gchar *utf8_filename = g_filename_to_utf8 (filename, -1,
-                                                           NULL, NULL, NULL);
+               gchar *display_name = g_filename_display_name (filename);
                g_set_error (error,
                             G_FILE_ERROR,
                             g_file_error_from_errno (errno),
                             _("Failed to close '%s' while writing image, all data may not have been saved: %s"),
-                            utf8_filename ? utf8_filename : "???",
+                            display_name,
                             g_strerror (errno));
-               g_free (utf8_filename);
+               g_free (display_name);
                return FALSE;
        }
        
        return TRUE;
 }
 
+#ifdef G_OS_WIN32
+
+#undef gdk_pixbuf_savev
+
+gboolean
+gdk_pixbuf_savev (GdkPixbuf  *pixbuf, 
+                  const char *filename, 
+                  const char *type,
+                  char      **option_keys,
+                  char      **option_values,
+                  GError    **error)
+{
+	char *utf8_filename;
+	gboolean retval;
+
+        g_return_val_if_fail (filename != NULL, FALSE);
+       
+	utf8_filename = g_locale_to_utf8 (filename, -1, NULL, NULL, error);
+
+	if (utf8_filename == NULL)
+		return FALSE;
+
+	retval = gdk_pixbuf_savev_utf8 (pixbuf, utf8_filename, type,
+					option_keys, option_values, error);
+
+	g_free (utf8_filename);
+
+	return retval;
+}
+
+#endif
+
 /**
  * gdk_pixbuf_save_to_callback:
  * @pixbuf: a #GdkPixbuf.
@@ -1550,7 +1816,7 @@ gdk_pixbuf_save_to_callback    (GdkPixbuf  *pixbuf,
  * @error: return location for error, or %NULL
  *
  * Saves pixbuf to a callback in format @type, which is currently "jpeg",
- * "png" or "ico".  If @error is set, %FALSE will be returned. See
+ * "png", "ico" or "bmp".  If @error is set, %FALSE will be returned. See
  * gdk_pixbuf_save_to_callback () for more details.
  *
  * Return value: whether an error was set
@@ -1596,7 +1862,7 @@ gdk_pixbuf_save_to_callbackv   (GdkPixbuf  *pixbuf,
  * @Varargs: list of key-value save options
  *
  * Saves pixbuf to a new buffer in format @type, which is currently "jpeg",
- * "png" or "ico".  This is a convenience function that uses
+ * "png", "ico" or "bmp".  This is a convenience function that uses
  * gdk_pixbuf_save_to_callback() to do the real work. Note that the buffer 
  * is not nul-terminated and may contain embedded  nuls.
  * If @error is set, %FALSE will be returned and @string will be set to
@@ -1684,7 +1950,7 @@ save_to_buffer_callback (const gchar *data,
  * @error: return location for error, or %NULL
  *
  * Saves pixbuf to a new buffer in format @type, which is currently "jpeg",
- * "png" or "ico".  See gdk_pixbuf_save_to_buffer() for more details.
+ * "png", "ico" or "bmp".  See gdk_pixbuf_save_to_buffer() for more details.
  *
  * Return value: whether an error was set
  *
@@ -1907,7 +2173,7 @@ gdk_pixbuf_format_set_disabled (GdkPixbufFormat *format,
 gchar*
 gdk_pixbuf_format_get_license (GdkPixbufFormat *format)
 {
-	g_return_val_if_fail (format != NULL, FALSE);
+	g_return_val_if_fail (format != NULL, NULL);
 
 	return g_strdup (format->license);
 }
@@ -1949,6 +2215,9 @@ gdk_pixbuf_get_formats (void)
 }
 
 
+#define __GDK_PIXBUF_IO_C__
+#include "gdk-pixbuf-aliasdef.c"
+
 
 
 
