@@ -5,9 +5,7 @@
 #include <gdk/gdktypes.h>
 #include <gdk/gdkevents.h>
 
-#ifdef __cplusplus
-extern "C" {
-#endif /* __cplusplus */
+G_BEGIN_DECLS
 
 typedef struct _GdkGeometry           GdkGeometry;
 typedef struct _GdkWindowAttr	      GdkWindowAttr;
@@ -260,6 +258,9 @@ struct _GdkWindowObject
   guint modal_hint : 1;
   
   guint destroyed : 2;
+
+  guint accept_focus : 1;
+  guint focus_on_map : 1;
   
   GdkEventMask event_mask;
 };
@@ -317,6 +318,10 @@ void          gdk_window_set_user_data         (GdkWindow     *window,
                                                 gpointer       user_data);
 void          gdk_window_set_override_redirect (GdkWindow     *window,
                                                 gboolean       override_redirect);
+void          gdk_window_set_accept_focus      (GdkWindow     *window,
+					        gboolean       accept_focus);
+void          gdk_window_set_focus_on_map      (GdkWindow     *window,
+					        gboolean       focus_on_map);
 void          gdk_window_add_filter            (GdkWindow     *window,
                                                 GdkFilterFunc  function,
                                                 gpointer       data);
@@ -409,6 +414,8 @@ void gdk_window_set_skip_taskbar_hint (GdkWindow *window,
                                        gboolean   skips_taskbar);
 void gdk_window_set_skip_pager_hint   (GdkWindow *window,
                                        gboolean   skips_pager);
+void gdk_window_set_urgency_hint      (GdkWindow *window,
+				       gboolean   urgent);
 
 void          gdk_window_set_geometry_hints (GdkWindow        *window,
 					     GdkGeometry      *geometry,
@@ -428,7 +435,7 @@ void          gdk_window_set_role          (GdkWindow       *window,
 void          gdk_window_set_transient_for (GdkWindow       *window, 
 					    GdkWindow       *parent);
 void	      gdk_window_set_background	 (GdkWindow	  *window,
-					  GdkColor	  *color);
+					  const GdkColor  *color);
 void	      gdk_window_set_back_pixmap (GdkWindow	  *window,
 					  GdkPixmap	  *pixmap,
 					  gboolean	   parent_relative);
@@ -449,7 +456,8 @@ gint	      gdk_window_get_origin	 (GdkWindow	  *window,
 					  gint		  *x,
 					  gint		  *y);
 
-#ifndef GDK_DISABLE_DEPRECATED
+#if !defined (GDK_DISABLE_DEPRECATED) || defined (GTK_COMPILATION)
+/* Used by gtk_handle_box_button_changed () */
 gboolean      gdk_window_get_deskrelative_origin (GdkWindow	  *window,
 					  gint		  *x,
 					  gint		  *y);
@@ -483,6 +491,7 @@ void	      gdk_window_set_icon_name	 (GdkWindow	  *window,
 					  const gchar	  *name);
 void	      gdk_window_set_group	 (GdkWindow	  *window, 
 					  GdkWindow	  *leader);
+GdkWindow*    gdk_window_get_group	 (GdkWindow	  *window);
 void	      gdk_window_set_decorations (GdkWindow	  *window,
 					  GdkWMDecoration  decorations);
 gboolean      gdk_window_get_decorations (GdkWindow       *window,
@@ -501,6 +510,10 @@ void          gdk_window_maximize        (GdkWindow       *window);
 void          gdk_window_unmaximize      (GdkWindow       *window);
 void          gdk_window_fullscreen      (GdkWindow       *window);
 void          gdk_window_unfullscreen    (GdkWindow       *window);
+void          gdk_window_set_keep_above  (GdkWindow       *window,
+                                          gboolean         setting);
+void          gdk_window_set_keep_below  (GdkWindow       *window,
+                                          gboolean         setting);
 
 void          gdk_window_register_dnd    (GdkWindow       *window);
 
@@ -551,6 +564,9 @@ void gdk_window_get_internal_paint_info (GdkWindow    *window,
 					 gint         *x_offset,
 					 gint         *y_offset);
 
+void gdk_window_enable_synchronized_configure (GdkWindow *window);
+void gdk_window_configure_finished            (GdkWindow *window);
+
 #ifndef GDK_MULTIHEAD_SAFE
 GdkPointerHooks *gdk_set_pointer_hooks (const GdkPointerHooks *new_hooks);   
 #endif /* GDK_MULTIHEAD_SAFE */
@@ -571,8 +587,6 @@ GdkWindow *gdk_get_default_root_window (void);
    gdk_draw_pixmap(drawable,gc,source_drawable,source_x,source_y,x,y,width,height)
 #endif /* GDK_DISABLE_DEPRECATED */
 
-#ifdef __cplusplus
-}
-#endif /* __cplusplus */
+G_END_DECLS
 
 #endif /* __GDK_WINDOW_H__ */
