@@ -24,9 +24,9 @@
  * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
  */
 
-#include "config.h"
+#include <config.h>
 
-#include <stdio.h>
+#include <glib/gprintf.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
@@ -43,12 +43,13 @@
 
 #include "gdk.h"
 
-#include "gdkprivate-x11.h"
+#include "gdkx.h"
+#include "gdkdisplay-x11.h"
 #include "gdkinternals.h"
+#include "gdkintl.h"
 #include "gdkregion-generic.h"
 #include "gdkinputprivate.h"
-
-#include <pango/pangox.h>
+#include "gdkalias.h"
 
 typedef struct _GdkPredicate  GdkPredicate;
 typedef struct _GdkErrorTrap  GdkErrorTrap;
@@ -83,188 +84,24 @@ static int	    gdk_x_io_error		 (Display     *display);
 
 /* Private variable declarations
  */
-static int gdk_initialized = 0;			    /* 1 if the library is initialized,
-						     * 0 otherwise.
-						     */
-
-static gint autorepeat;
-static gboolean gdk_synchronize = FALSE;
 static GSList *gdk_error_traps = NULL;               /* List of error traps */
 static GSList *gdk_error_trap_free_list = NULL;      /* Free list */
 
-/* Information about current pointer and keyboard grabs held by this
- * client. If gdk_pointer_xgrab_window or gdk_keyboard_xgrab_window
- * window is NULL, then the other associated fields are ignored
- */
-static GdkWindowObject *gdk_pointer_xgrab_window = NULL;
-static gulong gdk_pointer_xgrab_serial;
-static gboolean gdk_pointer_xgrab_owner_events;
-
-static GdkWindowObject *gdk_keyboard_xgrab_window = NULL;
-static gulong gdk_keyboard_xgrab_serial;
-static gboolean gdk_keyboard_xgrab_owner_events;
-
-GdkArgDesc _gdk_windowing_args[] = {
-  { "display",     GDK_ARG_STRING,   &_gdk_display_name,    (GdkArgFunc)NULL   },
-  { "sync",        GDK_ARG_BOOL,     &gdk_synchronize,     (GdkArgFunc)NULL   },
-  { "gxid-host",   GDK_ARG_STRING,   &_gdk_input_gxid_host, (GdkArgFunc)NULL   },
-  { "gxid-port",   GDK_ARG_INT,      &_gdk_input_gxid_port, (GdkArgFunc)NULL   },
+GOptionEntry _gdk_windowing_args[] = {
+  { "sync", 0, 0, G_OPTION_ARG_NONE, &_gdk_synchronize, 
+    /* Description of --sync in --help output */ N_("Make X calls synchronous"), NULL },
   { NULL }
 };
 
-/*
- * XLib internal connection handling
- */
-typedef struct _GdkInternalConnection GdkInternalConnection;
-
-struct _GdkInternalConnection
+void
+_gdk_windowing_init (void)
 {
-  gint	         fd;
-  GSource	*source;
-  Display	*display;
-};
-
-static gboolean
-process_internal_connection (GIOChannel  *gioc,
-			     GIOCondition cond,
-			     gpointer     data)
-{
-  GdkInternalConnection *connection = (GdkInternalConnection *)data;
-
-  GDK_THREADS_ENTER ();
-
-  XProcessInternalConnection ((Display*)connection->display, connection->fd);
-
-  GDK_THREADS_LEAVE ();
-
-  return TRUE;
-}
-
-static GdkInternalConnection *
-gdk_add_connection_handler (Display *display,
-			    guint    fd)
-{
-  GIOChannel *io_channel;
-  GdkInternalConnection *connection;
-
-  connection = g_new (GdkInternalConnection, 1);
-
-  connection->fd = fd;
-  connection->display = display;
-  
-  io_channel = g_io_channel_unix_new (fd);
-  
-  connection->source = g_io_create_watch (io_channel, G_IO_IN);
-  g_source_set_callback (connection->source,
-			 (GSourceFunc)process_internal_connection, connection, NULL);
-  g_source_attach (connection->source, NULL);
-  
-  g_io_channel_unref (io_channel);
-  
-  return connection;
-}
-
-static void
-gdk_remove_connection_handler (GdkInternalConnection *connection)
-{
-  g_source_destroy (connection->source);
-  g_free (connection);
-}
-
-static void
-gdk_internal_connection_watch (Display  *display,
-			       XPointer  arg,
-			       gint      fd,
-			       gboolean  opening,
-			       XPointer *watch_data)
-{
-  if (opening)
-    *watch_data = (XPointer)gdk_add_connection_handler (display, fd);
-  else
-    gdk_remove_connection_handler ((GdkInternalConnection *)*watch_data);
-}
-
-gboolean
-_gdk_windowing_init_check (int argc, char **argv)
-{
-  XKeyboardState keyboard_state;
-  XClassHint *class_hint;
-  gulong pid;
-
   _gdk_x11_initialize_locale ();
   
   XSetErrorHandler (gdk_x_error);
   XSetIOErrorHandler (gdk_x_io_error);
-  
-  gdk_display = XOpenDisplay (_gdk_display_name);
-  if (!gdk_display)
-    return FALSE;
 
-  XAddConnectionWatch (gdk_display, gdk_internal_connection_watch, NULL);
-  
-  if (gdk_synchronize)
-    XSynchronize (gdk_display, True);
-  
-  _gdk_screen = DefaultScreen (gdk_display);
-  _gdk_root_window = RootWindow (gdk_display, _gdk_screen);
-  
-  _gdk_leader_window = XCreateSimpleWindow(gdk_display, _gdk_root_window,
-					  10, 10, 10, 10, 0, 0 , 0);
-  class_hint = XAllocClassHint();
-  class_hint->res_name = g_get_prgname ();
-  class_hint->res_class = (char *)gdk_get_program_class ();
-  XmbSetWMProperties (gdk_display, _gdk_leader_window,
-                      NULL, NULL, argv, argc, 
-                      NULL, NULL, class_hint);
-  XFree (class_hint);
-
-  pid = getpid();
-  XChangeProperty (gdk_display, _gdk_leader_window,
-		   gdk_x11_get_xatom_by_name ("_NET_WM_PID"),
-		   XA_CARDINAL, 32,
-		   PropModeReplace,
-		   (guchar *)&pid, 1);
-  
   _gdk_selection_property = gdk_atom_intern ("GDK_SELECTION", FALSE);
-
-  XGetKeyboardControl (gdk_display, &keyboard_state);
-  autorepeat = keyboard_state.global_auto_repeat;
-
-#ifdef HAVE_XKB
-  {
-    gint xkb_major = XkbMajorVersion;
-    gint xkb_minor = XkbMinorVersion;
-    if (XkbLibraryVersion (&xkb_major, &xkb_minor))
-      {
-        xkb_major = XkbMajorVersion;
-        xkb_minor = XkbMinorVersion;
-	    
-        if (XkbQueryExtension (gdk_display, NULL, &_gdk_xkb_event_type, NULL,
-                               &xkb_major, &xkb_minor))
-          {
-	    Bool detectable_autorepeat_supported;
-
-            _gdk_use_xkb = TRUE;
-
-            XkbSelectEvents (gdk_display,
-                             XkbUseCoreKbd,
-                             XkbMapNotifyMask | XkbStateNotifyMask,
-                             XkbMapNotifyMask | XkbStateNotifyMask);
-
-	    XkbSetDetectableAutoRepeat (gdk_display,
-					True,
-					&detectable_autorepeat_supported);
-
-	    GDK_NOTE (MISC, g_message ("Detectable autorepeat %s.",
-				       detectable_autorepeat_supported ? "supported" : "not supported"));
-	    
-	    _gdk_have_xkb_autorepeat = detectable_autorepeat_supported;
-          }
-      }
-  }
-#endif
-  
-  return TRUE;
 }
 
 void
@@ -275,7 +112,7 @@ gdk_set_use_xshm (gboolean use_xshm)
 gboolean
 gdk_get_use_xshm (void)
 {
-  return _gdk_use_xshm;
+  return GDK_DISPLAY_X11 (gdk_display_get_default ())->use_xshm;
 }
 
 static GdkGrabStatus
@@ -357,8 +194,10 @@ gdk_pointer_grab (GdkWindow *	  window,
   if (!cursor)
     xcursor = None;
   else
-    xcursor = cursor_private->xcursor;
-  
+    {
+      _gdk_x11_cursor_update_theme (cursor);
+      xcursor = cursor_private->xcursor;
+    }
   
   xevent_mask = 0;
   for (i = 0; i < _gdk_nenvent_masks; i++)
@@ -397,61 +236,18 @@ gdk_pointer_grab (GdkWindow *	  window,
   
   if (return_val == GrabSuccess)
     {
-      gdk_pointer_xgrab_window = (GdkWindowObject *)window;
-      gdk_pointer_xgrab_serial = serial;
-      gdk_pointer_xgrab_owner_events = owner_events;
+      GdkDisplayX11 *display_x11 = GDK_DISPLAY_X11 (GDK_WINDOW_DISPLAY (window));
+      display_x11->pointer_xgrab_window = (GdkWindowObject *)window;
+      display_x11->pointer_xgrab_serial = serial;
+      display_x11->pointer_xgrab_owner_events = owner_events;
     }
 
   return gdk_x11_convert_grab_status (return_val);
 }
 
-/*
- *--------------------------------------------------------------
- * gdk_pointer_ungrab
- *
- *   Releases any pointer grab
- *
- * Arguments:
- *
- * Results:
- *
- * Side effects:
- *
- *--------------------------------------------------------------
- */
-
-void
-gdk_pointer_ungrab (guint32 time)
-{
-  _gdk_input_ungrab_pointer (time);
-  
-  XUngrabPointer (gdk_display, time);
-  gdk_pointer_xgrab_window = NULL;
-}
-
-/*
- *--------------------------------------------------------------
- * gdk_pointer_is_grabbed
- *
- *   Tell wether there is an active x pointer grab in effect
- *
- * Arguments:
- *
- * Results:
- *
- * Side effects:
- *
- *--------------------------------------------------------------
- */
-
-gboolean
-gdk_pointer_is_grabbed (void)
-{
-  return gdk_pointer_xgrab_window != NULL;
-}
-
 /**
  * gdk_pointer_grab_info_libgtk_only:
+ * @display: the #GdkDisplay for which to get the grab information
  * @grab_window: location to store current grab window
  * @owner_events: location to store boolean indicating whether
  *   the @owner_events flag to gdk_pointer_grab() was %TRUE.
@@ -463,15 +259,22 @@ gdk_pointer_is_grabbed (void)
  *  pointer grabbed.
  **/
 gboolean
-gdk_pointer_grab_info_libgtk_only (GdkWindow **grab_window,
+gdk_pointer_grab_info_libgtk_only (GdkDisplay *display,
+				   GdkWindow **grab_window,
 				   gboolean   *owner_events)
 {
-  if (gdk_pointer_xgrab_window)
+  GdkDisplayX11 *display_x11;
+  
+  g_return_val_if_fail (GDK_IS_DISPLAY (display), False);
+
+  display_x11 = GDK_DISPLAY_X11 (display);
+
+  if (display_x11->pointer_xgrab_window)
     {
       if (grab_window)
-        *grab_window = (GdkWindow *)gdk_pointer_xgrab_window;
+        *grab_window = (GdkWindow *)display_x11->pointer_xgrab_window;
       if (owner_events)
-        *owner_events = gdk_pointer_xgrab_owner_events;
+        *owner_events = display_x11->pointer_xgrab_owner_events;
 
       return TRUE;
     }
@@ -530,38 +333,18 @@ gdk_keyboard_grab (GdkWindow *	   window,
 
   if (return_val == GrabSuccess)
     {
-      gdk_keyboard_xgrab_window = (GdkWindowObject *)window;
-      gdk_keyboard_xgrab_serial = serial;
-      gdk_keyboard_xgrab_owner_events = owner_events;
+      GdkDisplayX11 *display_x11 = GDK_DISPLAY_X11 (gdk_drawable_get_display (window));
+      display_x11->keyboard_xgrab_window = (GdkWindowObject *)window;
+      display_x11->keyboard_xgrab_serial = serial;
+      display_x11->keyboard_xgrab_owner_events = owner_events;
     }
 
   return gdk_x11_convert_grab_status (return_val);
 }
 
-/*
- *--------------------------------------------------------------
- * gdk_keyboard_ungrab
- *
- *   Releases any keyboard grab
- *
- * Arguments:
- *
- * Results:
- *
- * Side effects:
- *
- *--------------------------------------------------------------
- */
-
-void
-gdk_keyboard_ungrab (guint32 time)
-{
-  XUngrabKeyboard (gdk_display, time);
-  gdk_keyboard_xgrab_window = NULL;
-}
-
 /**
  * gdk_keyboard_grab_info_libgtk_only:
+ * @display: the display for which to get the grab information
  * @grab_window: location to store current grab window
  * @owner_events: location to store boolean indicating whether
  *   the @owner_events flag to gdk_keyboard_grab() was %TRUE.
@@ -573,15 +356,22 @@ gdk_keyboard_ungrab (guint32 time)
  *  keyboard grabbed.
  **/
 gboolean
-gdk_keyboard_grab_info_libgtk_only (GdkWindow **grab_window,
+gdk_keyboard_grab_info_libgtk_only (GdkDisplay *display,
+				    GdkWindow **grab_window,
 				    gboolean   *owner_events)
 {
-  if (gdk_keyboard_xgrab_window)
+  GdkDisplayX11 *display_x11;
+  
+  g_return_val_if_fail (GDK_IS_DISPLAY (display), False);
+
+  display_x11 = GDK_DISPLAY_X11 (display);
+
+  if (display_x11->keyboard_xgrab_window)
     {
       if (grab_window)
-        *grab_window = (GdkWindow *)gdk_keyboard_xgrab_window;
+        *grab_window = (GdkWindow *)display_x11->keyboard_xgrab_window;
       if (owner_events)
-        *owner_events = gdk_keyboard_xgrab_owner_events;
+        *owner_events = display_x11->keyboard_xgrab_owner_events;
 
       return TRUE;
     }
@@ -603,30 +393,33 @@ void
 _gdk_xgrab_check_unmap (GdkWindow *window,
 			gulong     serial)
 {
-  if (gdk_pointer_xgrab_window && serial >= gdk_pointer_xgrab_serial)
+  GdkDisplayX11 *display_x11 = GDK_DISPLAY_X11 (gdk_drawable_get_display (window));
+  
+  if (display_x11->pointer_xgrab_window && 
+      serial >= display_x11->pointer_xgrab_serial)
     {
       GdkWindowObject *private = GDK_WINDOW_OBJECT (window);
-      GdkWindowObject *tmp = gdk_pointer_xgrab_window;
-      
+      GdkWindowObject *tmp = display_x11->pointer_xgrab_window;
 
       while (tmp && tmp != private)
 	tmp = tmp->parent;
 
       if (tmp)
-	gdk_pointer_xgrab_window = NULL;  
+	display_x11->pointer_xgrab_window = NULL;  
     }
 
-  if (gdk_keyboard_xgrab_window && serial >= gdk_keyboard_xgrab_serial)
+  if (display_x11->keyboard_xgrab_window &&
+      serial >= display_x11->keyboard_xgrab_serial)
     {
       GdkWindowObject *private = GDK_WINDOW_OBJECT (window);
-      GdkWindowObject *tmp = gdk_keyboard_xgrab_window;
+      GdkWindowObject *tmp = display_x11->keyboard_xgrab_window;
       
 
       while (tmp && tmp != private)
 	tmp = tmp->parent;
 
       if (tmp)
-	gdk_keyboard_xgrab_window = NULL;  
+	display_x11->keyboard_xgrab_window = NULL;  
     }
 }
 
@@ -640,162 +433,49 @@ _gdk_xgrab_check_unmap (GdkWindow *window,
 void
 _gdk_xgrab_check_destroy (GdkWindow *window)
 {
-  if ((GdkWindowObject *)window == gdk_pointer_xgrab_window)
-    gdk_pointer_xgrab_window = NULL;
+  GdkDisplayX11 *display_x11 = GDK_DISPLAY_X11 (gdk_drawable_get_display (window));
+  
+  if ((GdkWindowObject *)window == display_x11->pointer_xgrab_window)
+     display_x11->pointer_xgrab_window = NULL;
 
-  if ((GdkWindowObject *)window == gdk_keyboard_xgrab_window)
-    gdk_keyboard_xgrab_window = NULL;
+  if ((GdkWindowObject *)window ==  display_x11->keyboard_xgrab_window)
+     display_x11->keyboard_xgrab_window = NULL;
 }
 
-/*
- *--------------------------------------------------------------
- * gdk_screen_width
- *
- *   Return the width of the screen.
- *
- * Arguments:
- *
- * Results:
- *
- * Side effects:
- *
- *--------------------------------------------------------------
- */
-
-gint
-gdk_screen_width (void)
-{
-  gint return_val;
-  
-  return_val = DisplayWidth (gdk_display, _gdk_screen);
-  
-  return return_val;
-}
-
-/*
- *--------------------------------------------------------------
- * gdk_screen_height
- *
- *   Return the height of the screen.
- *
- * Arguments:
- *
- * Results:
- *
- * Side effects:
- *
- *--------------------------------------------------------------
- */
-
-gint
-gdk_screen_height (void)
-{
-  gint return_val;
-  
-  return_val = DisplayHeight (gdk_display, _gdk_screen);
-  
-  return return_val;
-}
-
-/*
- *--------------------------------------------------------------
- * gdk_screen_width_mm
- *
- *   Return the width of the screen in millimeters.
- *
- * Arguments:
- *
- * Results:
- *
- * Side effects:
- *
- *--------------------------------------------------------------
- */
-
-gint
-gdk_screen_width_mm (void)
-{
-  gint return_val;
-  
-  return_val = DisplayWidthMM (gdk_display, _gdk_screen);
-  
-  return return_val;
-}
-
-/*
- *--------------------------------------------------------------
- * gdk_screen_height
- *
- *   Return the height of the screen in millimeters.
- *
- * Arguments:
- *
- * Results:
- *
- * Side effects:
- *
- *--------------------------------------------------------------
- */
-
-gint
-gdk_screen_height_mm (void)
-{
-  gint return_val;
-  
-  return_val = DisplayHeightMM (gdk_display, _gdk_screen);
-  
-  return return_val;
-}
-
-/*
- *--------------------------------------------------------------
- * gdk_set_sm_client_id
- *
- *   Set the SM_CLIENT_ID property on the WM_CLIENT_LEADER window
- *   so that the window manager can save our state using the
- *   X11R6 ICCCM session management protocol. A NULL value should 
- *   be set following disconnection from the session manager to
- *   remove the SM_CLIENT_ID property.
- *
- * Arguments:
- * 
- *   "sm_client_id" specifies the client id assigned to us by the
- *   session manager or NULL to remove the property.
- *
- * Results:
- *
- * Side effects:
- *
- *--------------------------------------------------------------
- */
-
 void
-gdk_set_sm_client_id (const gchar* sm_client_id)
+_gdk_windowing_display_set_sm_client_id (GdkDisplay  *display,
+					 const gchar *sm_client_id)
 {
+  GdkDisplayX11 *display_x11 = GDK_DISPLAY_X11 (display);
+
+  if (display->closed)
+    return;
+
   if (sm_client_id && strcmp (sm_client_id, ""))
     {
-      XChangeProperty (gdk_display, _gdk_leader_window,
-	   	       gdk_x11_get_xatom_by_name ("SM_CLIENT_ID"),
-		       XA_STRING, 8, PropModeReplace,
-		       sm_client_id, strlen(sm_client_id));
+      XChangeProperty (display_x11->xdisplay, display_x11->leader_window,
+		       gdk_x11_get_xatom_by_name_for_display (display, "SM_CLIENT_ID"),
+		       XA_STRING, 8, PropModeReplace, sm_client_id,
+		       strlen (sm_client_id));
     }
   else
-     XDeleteProperty (gdk_display, _gdk_leader_window,
-	   	      gdk_x11_get_xatom_by_name ("SM_CLIENT_ID"));
-}
-
-void
-gdk_beep (void)
-{
-  XBell(gdk_display, 0);
+    XDeleteProperty (display_x11->xdisplay, display_x11->leader_window,
+		     gdk_x11_get_xatom_by_name_for_display (display, "SM_CLIENT_ID"));
 }
 
+/* Close all open displays
+ */
 void
 _gdk_windowing_exit (void)
 {
-  pango_x_shutdown_display (gdk_display);
-  
-  XCloseDisplay (gdk_display);
+  GSList *tmp_list = _gdk_displays;
+    
+  while (tmp_list)
+    {
+      XCloseDisplay (GDK_DISPLAY_XDISPLAY (tmp_list->data));
+      
+      tmp_list = tmp_list->next;
+  }
 }
 
 /*
@@ -853,9 +533,9 @@ gdk_x_error (Display	 *display,
 #ifdef G_ENABLE_DEBUG	  
 	  g_error ("%s", msg);
 #else /* !G_ENABLE_DEBUG */
-	  fprintf (stderr, "%s\n", msg);
+	  g_fprintf (stderr, "%s\n", msg);
 
-	  exit(1);
+	  exit (1);
 #endif /* G_ENABLE_DEBUG */
 	}
       _gdk_error_code = error->error_code;
@@ -894,24 +574,21 @@ gdk_x_io_error (Display *display)
    */
   if (errno == EPIPE)
     {
-      fprintf (stderr,
+      g_fprintf (stderr,
                "The application '%s' lost its connection to the display %s;\n"
                "most likely the X server was shut down or you killed/destroyed\n"
                "the application.\n",
                g_get_prgname (),
-               gdk_display ? DisplayString (gdk_display) : gdk_get_display());
+               display ? DisplayString (display) : gdk_get_display_arg_name ());
     }
   else
     {
-      fprintf (stderr, "%s: Fatal IO error %d (%s) on X server %s.\n",
+      g_fprintf (stderr, "%s: Fatal IO error %d (%s) on X server %s.\n",
                g_get_prgname (),
 	       errno, g_strerror (errno),
-	       gdk_display ? DisplayString (gdk_display) : gdk_get_display());
+	       display ? DisplayString (display) : gdk_get_display_arg_name ());
     }
 
-  /* Disable the atexit shutdown for GDK */
-  gdk_initialized = 0;
-  
   exit(1);
 }
 
@@ -991,18 +668,40 @@ gdk_error_trap_pop (void)
 gchar *
 gdk_get_display (void)
 {
-  return (gchar *)XDisplayName (_gdk_display_name);
+  return g_strdup (gdk_display_get_name (gdk_display_get_default ()));
 }
 
+/**
+ * _gdk_send_xevent:
+ * @display: #GdkDisplay which @window is on
+ * @window: window ID to which to send the event
+ * @propagate: %TRUE if the event should be propagated if the target window
+ *             doesn't handle it.
+ * @event_mask: event mask to match against, or 0 to send it to @window
+ *              without regard to event masks.
+ * @event_send: #XEvent to send
+ * 
+ * Send an event, like XSendEvent(), but trap errors and check
+ * the result.
+ * 
+ * Return value: %TRUE if sending the event succeeded.
+ **/
 gint 
-gdk_send_xevent (Window window, gboolean propagate, glong event_mask,
-		 XEvent *event_send)
+_gdk_send_xevent (GdkDisplay *display,
+		  Window      window, 
+		  gboolean    propagate, 
+		  glong       event_mask,
+		  XEvent     *event_send)
 {
   gboolean result;
-  
+
+  if (display->closed)
+    return FALSE;
+
   gdk_error_trap_push ();
-  result = XSendEvent (gdk_display, window, propagate, event_mask, event_send);
-  XSync (gdk_display, False);
+  result = XSendEvent (GDK_DISPLAY_XDISPLAY (display), window, 
+		       propagate, event_mask, event_send);
+  XSync (GDK_DISPLAY_XDISPLAY (display), False);
   
   return result && gdk_error_trap_pop() == Success;
 }
@@ -1030,24 +729,30 @@ _gdk_region_get_xrectangles (GdkRegion   *region,
   *n_rects = region->numRects;
 }
 
-/* FIXME put in GdkDisplay */
-static gint grab_count = 0;
+/**
+ * gdk_x11_grab_server:
+ * 
+ * Call gdk_x11_display_grab() on the default display. 
+ * To ungrab the server again, use gdk_x11_ungrab_server(). 
+ *
+ * gdk_x11_grab_server()/gdk_x11_ungrab_server() calls can be nested.
+ **/ 
 void
 gdk_x11_grab_server (void)
-{ 
-  if (grab_count == 0)
-    XGrabServer (gdk_display);
-  ++grab_count;
+{
+  gdk_x11_display_grab (gdk_display_get_default ());
 }
 
+/**
+ * gdk_x11_ungrab_server:
+ *
+ * Ungrab the default display after it has been grabbed with 
+ * gdk_x11_grab_server(). 
+ **/
 void
 gdk_x11_ungrab_server (void)
 {
-  g_return_if_fail (grab_count > 0);
-  
-  --grab_count;
-  if (grab_count == 0)
-    XUngrabServer (gdk_display);
+  gdk_x11_display_ungrab (gdk_display_get_default ());
 }
 
 /**
@@ -1056,23 +761,43 @@ gdk_x11_ungrab_server (void)
  * Gets the default GTK+ screen number.
  * 
  * Return value: returns the screen number specified by
- *   the --display command line option on the DISPLAY environment
- *   variable gdk_init() calls XOpenDisplay().
+ *   the --display command line option or the DISPLAY environment
+ *   variable when gdk_init() calls XOpenDisplay().
  **/
 gint
 gdk_x11_get_default_screen (void)
 {
-  return _gdk_screen;
+  return gdk_screen_get_number (gdk_screen_get_default ());
 }
 
+/**
+ * gdk_x11_get_default_root_xwindow:
+ * 
+ * Gets the root window of the default screen 
+ * (see gdk_x11_get_default_screen()).  
+ * 
+ * Return value: an Xlib <type>Window</type>.
+ **/
 Window
 gdk_x11_get_default_root_xwindow (void)
 {
-  return _gdk_root_window;
+  return GDK_SCREEN_XROOTWIN (gdk_screen_get_default ());
 }
 
+/**
+ * gdk_x11_get_default_xdisplay:
+ * 
+ * Gets the default GTK+ display.
+ * 
+ * Return value: the Xlib <type>Display*</type> for the display
+ * specified in the <option>--display</option> command line option 
+ * or the <envar>DISPLAY</envar> environment variable.
+ **/
 Display *
 gdk_x11_get_default_xdisplay (void)
 {
-  return gdk_display;
+  return GDK_DISPLAY_XDISPLAY (gdk_display_get_default ());
 }
+
+#define __GDK_MAIN_X11_C__
+#include "gdkaliasdef.c"
