@@ -23,11 +23,11 @@
 #include <config.h>
 #include <string.h>
 
-#include "gtkalias.h"
 #include "gtkclipboard.h"
 #include "gtkinvisible.h"
 #include "gtkmain.h"
 #include "gtkmarshalers.h"
+#include "gtkalias.h"
 
 #ifdef GDK_WINDOWING_X11
 #include "x11/gdkx.h"
@@ -678,7 +678,7 @@ clipboard_unset (GtkClipboard *clipboard)
    */
   if (old_have_owner &&
       old_n_storable_targets != -1)
-    g_object_unref (clipboard->user_data);
+    g_object_unref (old_data);
 }
 
 /**
@@ -1609,6 +1609,7 @@ gtk_clipboard_set_can_store (GtkClipboard         *clipboard,
   };
   
   g_return_if_fail (GTK_IS_CLIPBOARD (clipboard));
+  g_return_if_fail (n_targets >= 0);
 
   if (clipboard->selection != GDK_SELECTION_CLIPBOARD)
     return;
@@ -1731,3 +1732,6 @@ _gtk_clipboard_store_all (void)
   g_slist_free (displays);
   
 }
+
+#define __GTK_CLIPBOARD_C__
+#include "gtkaliasdef.c"
