@@ -29,16 +29,26 @@
 #include <config.h>
 #include "gdk/gdkkeysyms.h"
 #include "gtkbindings.h"
+#include "gtkkeyhash.h"
 #include "gtkmain.h"
 #include "gtkmarshalers.h"
 #include "gtkmenubar.h"
 #include "gtkmenuitem.h"
 #include "gtkmenushell.h"
+#include "gtkmnemonichash.h"
 #include "gtktearoffmenuitem.h"
 #include "gtkwindow.h"
+#include "gtkprivate.h"
+#include "gtkintl.h"
+#include "gtkalias.h"
 
 #define MENU_SHELL_TIMEOUT   500
 
+#define PACK_DIRECTION(m)                                 \
+   (GTK_IS_MENU_BAR (m)                                   \
+     ? gtk_menu_bar_get_pack_direction (GTK_MENU_BAR (m)) \
+     : GTK_PACK_DIRECTION_LTR)
+
 enum {
   DEACTIVATE,
   SELECTION_DONE,
@@ -49,6 +59,11 @@ enum {
   LAST_SIGNAL
 };
 
+enum {
+  PROP_0,
+  PROP_TAKE_FOCUS
+};
+
 typedef void (*GtkMenuShellSignal1) (GtkObject           *object,
 				     GtkMenuDirectionType arg1,
 				     gpointer             data);
@@ -101,6 +116,10 @@ typedef void (*GtkMenuShellSignal2) (GtkObject *object,
  *       - For 'child', if there is no child, then current is
  *         moved to the next item in the parent.
  *
+ *    Note that the above explanation of ::move_current was written
+ *    before menus and menubars had support for RTL flipping and
+ *    different packing directions, and therefore only applies for
+ *    when text direction and packing direction are both left-to-right.
  * 
  *  ::activate_current (GBoolean *force_hide)
  *     Activate the current item. If 'force_hide' is true, hide
@@ -111,9 +130,30 @@ typedef void (*GtkMenuShellSignal2) (GtkObject *object,
  *     Cancels the current selection
  */
 
+#define GTK_MENU_SHELL_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GTK_TYPE_MENU_SHELL, GtkMenuShellPrivate))
+
+typedef struct _GtkMenuShellPrivate GtkMenuShellPrivate;
+
+struct _GtkMenuShellPrivate
+{
+  GtkMnemonicHash *mnemonic_hash;
+  GtkKeyHash *key_hash;
+
+  gboolean take_focus;
+};
+
 static void gtk_menu_shell_class_init        (GtkMenuShellClass *klass);
 static void gtk_menu_shell_init              (GtkMenuShell      *menu_shell);
+static void gtk_menu_shell_set_property      (GObject           *object,
+                                              guint              prop_id,
+                                              const GValue      *value,
+                                              GParamSpec        *pspec);
+static void gtk_menu_shell_get_property      (GObject           *object,
+                                              guint              prop_id,
+                                              GValue            *value,
+                                              GParamSpec        *pspec);
 static void gtk_menu_shell_realize           (GtkWidget         *widget);
+static void gtk_menu_shell_finalize          (GObject           *object);
 static gint gtk_menu_shell_button_press      (GtkWidget         *widget,
 					      GdkEventButton    *event);
 static gint gtk_menu_shell_button_release    (GtkWidget         *widget,
@@ -124,6 +164,8 @@ static gint gtk_menu_shell_enter_notify      (GtkWidget         *widget,
 					      GdkEventCrossing  *event);
 static gint gtk_menu_shell_leave_notify      (GtkWidget         *widget,
 					      GdkEventCrossing  *event);
+static void gtk_menu_shell_screen_changed    (GtkWidget         *widget,
+					      GdkScreen         *previous_screen);
 static void gtk_menu_shell_add               (GtkContainer      *container,
 					      GtkWidget         *widget);
 static void gtk_menu_shell_remove            (GtkContainer      *container,
@@ -143,7 +185,7 @@ static GtkWidget *gtk_menu_shell_get_item    (GtkMenuShell      *menu_shell,
 static GType    gtk_menu_shell_child_type  (GtkContainer      *container);
 static void gtk_menu_shell_real_select_item  (GtkMenuShell      *menu_shell,
 					      GtkWidget         *menu_item);
-static void gtk_menu_shell_select_submenu_first (GtkMenuShell   *menu_shell); 
+static gboolean gtk_menu_shell_select_submenu_first (GtkMenuShell   *menu_shell); 
 
 static void gtk_real_menu_shell_move_current (GtkMenuShell      *menu_shell,
 					      GtkMenuDirectionType direction);
@@ -153,6 +195,10 @@ static void gtk_real_menu_shell_cancel           (GtkMenuShell      *menu_shell)
 static void gtk_real_menu_shell_cycle_focus      (GtkMenuShell      *menu_shell,
 						  GtkDirectionType   dir);
 
+static void     gtk_menu_shell_reset_key_hash    (GtkMenuShell *menu_shell);
+static gboolean gtk_menu_shell_activate_mnemonic (GtkMenuShell *menu_shell,
+						  GdkEventKey  *event);
+
 static GtkContainerClass *parent_class = NULL;
 static guint menu_shell_signals[LAST_SIGNAL] = { 0 };
 
@@ -201,12 +247,17 @@ gtk_menu_shell_class_init (GtkMenuShellClass *klass)
 
   parent_class = g_type_class_peek_parent (klass);
 
+  object_class->set_property = gtk_menu_shell_set_property;
+  object_class->get_property = gtk_menu_shell_get_property;
+  object_class->finalize = gtk_menu_shell_finalize;
+
   widget_class->realize = gtk_menu_shell_realize;
   widget_class->button_press_event = gtk_menu_shell_button_press;
   widget_class->button_release_event = gtk_menu_shell_button_release;
   widget_class->key_press_event = gtk_menu_shell_key_press;
   widget_class->enter_notify_event = gtk_menu_shell_enter_notify;
   widget_class->leave_notify_event = gtk_menu_shell_leave_notify;
+  widget_class->screen_changed = gtk_menu_shell_screen_changed;
 
   container_class->add = gtk_menu_shell_add;
   container_class->remove = gtk_menu_shell_remove;
@@ -307,6 +358,25 @@ gtk_menu_shell_class_init (GtkMenuShellClass *klass)
 				GDK_F10, GDK_SHIFT_MASK,
 				"cycle_focus", 1,
                                 GTK_TYPE_DIRECTION_TYPE, GTK_DIR_TAB_BACKWARD);
+
+  /**
+   * GtkMenuShell:take-focus:
+   *
+   * A boolean that determines whether the menu and its submenus grab the
+   * keyboard focus. See gtk_menu_shell_set_take_focus() and
+   * gtk_menu_shell_get_take_focus().
+   *
+   * Since: 2.8
+   **/
+  g_object_class_install_property (object_class,
+                                   PROP_TAKE_FOCUS,
+                                   g_param_spec_boolean ("take-focus",
+							 P_("Take Focus"),
+							 P_("A boolean that determines whether the menu grabs the keyboard focus"),
+							 TRUE,
+							 GTK_PARAM_READWRITE));
+
+  g_type_class_add_private (object_class, sizeof (GtkMenuShellPrivate));
 }
 
 static GType
@@ -318,6 +388,8 @@ gtk_menu_shell_child_type (GtkContainer     *container)
 static void
 gtk_menu_shell_init (GtkMenuShell *menu_shell)
 {
+  GtkMenuShellPrivate *priv = GTK_MENU_SHELL_GET_PRIVATE (menu_shell);
+
   menu_shell->children = NULL;
   menu_shell->active_menu_item = NULL;
   menu_shell->parent_menu_shell = NULL;
@@ -326,8 +398,65 @@ gtk_menu_shell_init (GtkMenuShell *menu_shell)
   menu_shell->have_xgrab = FALSE;
   menu_shell->button = 0;
   menu_shell->activate_time = 0;
+
+  priv->mnemonic_hash = NULL;
+  priv->key_hash = NULL;
+  priv->take_focus = TRUE;
 }
 
+static void
+gtk_menu_shell_set_property (GObject      *object,
+                             guint         prop_id,
+                             const GValue *value,
+                             GParamSpec   *pspec)
+{
+  GtkMenuShell *menu_shell = GTK_MENU_SHELL (object);
+
+  switch (prop_id)
+    {
+    case PROP_TAKE_FOCUS:
+      gtk_menu_shell_set_take_focus (menu_shell, g_value_get_boolean (value));
+      break;
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+      break;
+    }
+}
+
+static void
+gtk_menu_shell_get_property (GObject     *object,
+                             guint        prop_id,
+                             GValue      *value,
+                             GParamSpec  *pspec)
+{
+  GtkMenuShell *menu_shell = GTK_MENU_SHELL (object);
+
+  switch (prop_id)
+    {
+    case PROP_TAKE_FOCUS:
+      g_value_set_boolean (value, gtk_menu_shell_get_take_focus (menu_shell));
+      break;
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+      break;
+    }
+}
+
+static void
+gtk_menu_shell_finalize (GObject *object)
+{
+  GtkMenuShell *menu_shell = GTK_MENU_SHELL (object);
+  GtkMenuShellPrivate *priv = GTK_MENU_SHELL_GET_PRIVATE (menu_shell);
+
+  if (priv->mnemonic_hash)
+    _gtk_mnemonic_hash_free (priv->mnemonic_hash);
+  if (priv->key_hash)
+    _gtk_key_hash_free (priv->key_hash);
+
+  G_OBJECT_CLASS (parent_class)->finalize (object);
+}
+
+
 void
 gtk_menu_shell_append (GtkMenuShell *menu_shell,
 		       GtkWidget    *child)
@@ -454,9 +583,10 @@ gtk_menu_shell_button_press (GtkWidget      *widget,
 	      (menu_item != menu_shell->active_menu_item))
 	    {
 	      if (GTK_MENU_SHELL_GET_CLASS (menu_shell)->submenu_placement == GTK_TOP_BOTTOM)
-		g_object_set_data (G_OBJECT (menu_shell),
-				   "gtk-menushell-just-activated",
-				   GUINT_TO_POINTER (1));
+		{
+		  menu_shell->activate_time = event->time;
+		}
+      
 	      gtk_menu_shell_select_item (menu_shell, menu_item);
 	    }
 	}
@@ -488,8 +618,6 @@ gtk_menu_shell_button_release (GtkWidget      *widget,
   menu_shell = GTK_MENU_SHELL (widget);
   if (menu_shell->active)
     {
-      gboolean deactivate_immediately = FALSE;
-
       if (menu_shell->button && (event->button != menu_shell->button))
 	{
 	  menu_shell->button = 0;
@@ -502,23 +630,8 @@ gtk_menu_shell_button_release (GtkWidget      *widget,
 
       deactivate = TRUE;
 
-      if (menu_item
-	  && GTK_MENU_SHELL_GET_CLASS (menu_shell)->submenu_placement == GTK_TOP_BOTTOM)
-	{
-	  if (g_object_get_data (G_OBJECT (menu_shell), "gtk-menushell-just-activated"))
-	    g_object_set_data (G_OBJECT (menu_shell), "gtk-menushell-just-activated", NULL);
-	  else
-	    deactivate_immediately = TRUE;
-	}
-
       if ((event->time - menu_shell->activate_time) > MENU_SHELL_TIMEOUT)
 	{
-	  if (deactivate_immediately)
-	    {
-	      gtk_menu_shell_deactivate (menu_shell);
-	      return TRUE;
-	    }
-	    
 	  if (menu_item && (menu_shell->active_menu_item == menu_item) &&
 	      _gtk_menu_item_is_selectable (menu_item))
 	    {
@@ -527,9 +640,18 @@ gtk_menu_shell_button_release (GtkWidget      *widget,
 		  gtk_menu_shell_activate_item (menu_shell, menu_item, TRUE);
 		  return TRUE;
 		}
+	      else if (GTK_MENU_SHELL_GET_CLASS (menu_shell)->submenu_placement != GTK_TOP_BOTTOM)
+		{
+		  gtk_menu_item_select (GTK_MENU_ITEM (menu_item));
+		  return TRUE;
+		}
+	    }
+	  else if (menu_item &&
+		   !_gtk_menu_item_is_selectable (menu_item) &&
+		   GTK_MENU_SHELL_GET_CLASS (menu_shell)->submenu_placement != GTK_TOP_BOTTOM)
+	    {
+	      deactivate = FALSE;
 	    }
-	  else if (menu_item && !_gtk_menu_item_is_selectable (menu_item))
-	    deactivate = FALSE;
 	  else if (menu_shell->parent_menu_shell)
 	    {
 	      menu_shell->active = TRUE;
@@ -539,8 +661,11 @@ gtk_menu_shell_button_release (GtkWidget      *widget,
 
 	  /* If we ended up on an item with a submenu, leave the menu up.
 	   */
-	  if (menu_item && (menu_shell->active_menu_item == menu_item))
-	    deactivate = FALSE;
+	  if (menu_item && (menu_shell->active_menu_item == menu_item) &&
+	      GTK_MENU_SHELL_GET_CLASS (menu_shell)->submenu_placement != GTK_TOP_BOTTOM)
+	    {
+	      deactivate = FALSE;
+	    }
 	}
       else /* a very fast press-release */
 	{
@@ -566,11 +691,10 @@ gtk_menu_shell_button_release (GtkWidget      *widget,
 }
 
 static gint
-gtk_menu_shell_key_press (GtkWidget	*widget,
+gtk_menu_shell_key_press (GtkWidget   *widget,
 			  GdkEventKey *event)
 {
   GtkMenuShell *menu_shell;
-  GtkWidget *toplevel;
   
   g_return_val_if_fail (GTK_IS_MENU_SHELL (widget), FALSE);
   g_return_val_if_fail (event != NULL, FALSE);
@@ -583,12 +707,7 @@ gtk_menu_shell_key_press (GtkWidget	*widget,
   if (gtk_bindings_activate_event (GTK_OBJECT (widget), event))
     return TRUE;
 
-  toplevel = gtk_widget_get_toplevel (widget);
-  if (GTK_IS_WINDOW (toplevel) &&
-      gtk_window_activate_key (GTK_WINDOW (toplevel), event))
-    return TRUE;
-
-  return FALSE;
+  return gtk_menu_shell_activate_mnemonic (menu_shell, event);
 }
 
 static gint
@@ -676,6 +795,13 @@ gtk_menu_shell_leave_notify (GtkWidget        *widget,
   return TRUE;
 }
 
+static void
+gtk_menu_shell_screen_changed (GtkWidget *widget,
+			       GdkScreen *previous_screen)
+{
+  gtk_menu_shell_reset_key_hash (GTK_MENU_SHELL (widget));
+}
+
 static void
 gtk_menu_shell_add (GtkContainer *container,
 		    GtkWidget    *widget)
@@ -745,6 +871,7 @@ gtk_real_menu_shell_deactivate (GtkMenuShell *menu_shell)
     {
       menu_shell->button = 0;
       menu_shell->active = FALSE;
+      menu_shell->activate_time = 0;
 
       if (menu_shell->active_menu_item)
 	{
@@ -831,14 +958,20 @@ static void
 gtk_menu_shell_real_select_item (GtkMenuShell *menu_shell,
 				 GtkWidget    *menu_item)
 {
+  GtkPackDirection pack_dir = PACK_DIRECTION (menu_shell);
+
   gtk_menu_shell_deselect (menu_shell);
 
   if (!_gtk_menu_item_is_selectable (menu_item))
     return;
 
   menu_shell->active_menu_item = menu_item;
-  _gtk_menu_item_set_placement (GTK_MENU_ITEM (menu_shell->active_menu_item),
-			       GTK_MENU_SHELL_GET_CLASS (menu_shell)->submenu_placement);
+  if (pack_dir == GTK_PACK_DIRECTION_TTB || pack_dir == GTK_PACK_DIRECTION_BTT)
+    _gtk_menu_item_set_placement (GTK_MENU_ITEM (menu_shell->active_menu_item),
+				  GTK_LEFT_RIGHT);
+  else
+    _gtk_menu_item_set_placement (GTK_MENU_ITEM (menu_shell->active_menu_item),
+				  GTK_MENU_SHELL_GET_CLASS (menu_shell)->submenu_placement);
   gtk_menu_item_select (GTK_MENU_ITEM (menu_shell->active_menu_item));
 
   /* This allows the bizarre radio buttons-with-submenus-display-history
@@ -875,6 +1008,7 @@ gtk_menu_shell_activate_item (GtkMenuShell      *menu_shell,
     deactivate = GTK_MENU_ITEM_GET_CLASS (menu_item)->hide_on_activate;
 
   g_object_ref (menu_shell);
+  g_object_ref (menu_item);
 
   if (deactivate)
     {
@@ -907,6 +1041,7 @@ gtk_menu_shell_activate_item (GtkMenuShell      *menu_shell,
   g_slist_free (shells);
 
   g_object_unref (menu_shell);
+  g_object_unref (menu_item);
 }
 
 /* Distance should be +/- 1 */
@@ -1019,7 +1154,7 @@ _gtk_menu_shell_select_last (GtkMenuShell *menu_shell,
     gtk_menu_shell_select_item (menu_shell, to_select);
 }
 
-static void
+static gboolean
 gtk_menu_shell_select_submenu_first (GtkMenuShell     *menu_shell)
 {
   GtkMenuItem *menu_item;
@@ -1027,7 +1162,14 @@ gtk_menu_shell_select_submenu_first (GtkMenuShell     *menu_shell)
   menu_item = GTK_MENU_ITEM (menu_shell->active_menu_item); 
   
   if (menu_item->submenu)
-    gtk_menu_shell_select_first (GTK_MENU_SHELL (menu_item->submenu), TRUE);
+    {
+      _gtk_menu_item_popup_submenu (GTK_WIDGET (menu_item));
+      gtk_menu_shell_select_first (GTK_MENU_SHELL (menu_item->submenu), TRUE);
+      if (GTK_MENU_SHELL (menu_item->submenu)->active_menu_item)
+	return TRUE;
+    }
+
+  return FALSE;
 }
 
 static void
@@ -1041,29 +1183,7 @@ gtk_real_menu_shell_move_current (GtkMenuShell      *menu_shell,
 
   if (menu_shell->parent_menu_shell)
     parent_menu_shell = GTK_MENU_SHELL (menu_shell->parent_menu_shell);
-  
-  if (gtk_widget_get_direction (GTK_WIDGET (menu_shell)) == GTK_TEXT_DIR_RTL)
-    {
-      switch (direction) 
-	{
-	case GTK_MENU_DIR_PARENT:
-	  direction = GTK_MENU_DIR_CHILD;
-	  break;
-	case GTK_MENU_DIR_CHILD:
-	  direction = GTK_MENU_DIR_PARENT;
-	  break;
-	case GTK_MENU_DIR_PREV:
-	  if (GTK_MENU_SHELL_GET_CLASS (menu_shell)->submenu_placement == GTK_TOP_BOTTOM)
-	    direction = GTK_MENU_DIR_NEXT;
-	  break;
-	case GTK_MENU_DIR_NEXT:
-	  if (GTK_MENU_SHELL_GET_CLASS (menu_shell)->submenu_placement == GTK_TOP_BOTTOM)
-	    direction = GTK_MENU_DIR_PREV;
-	  break;
-	default: ;
-	}
-    }
-  
+
   switch (direction)
     {
     case GTK_MENU_DIR_PARENT:
@@ -1074,7 +1194,10 @@ gtk_real_menu_shell_move_current (GtkMenuShell      *menu_shell,
 	    gtk_menu_shell_deselect (menu_shell);
 	  else 
 	    {
-	      gtk_menu_shell_move_selected (parent_menu_shell, -1);
+	      if (PACK_DIRECTION (parent_menu_shell) == GTK_PACK_DIRECTION_LTR)
+		gtk_menu_shell_move_selected (parent_menu_shell, -1);
+	      else
+		gtk_menu_shell_move_selected (parent_menu_shell, 1);
 	      gtk_menu_shell_select_submenu_first (parent_menu_shell); 
 	    }
 	}
@@ -1099,28 +1222,31 @@ gtk_real_menu_shell_move_current (GtkMenuShell      *menu_shell,
 	  _gtk_menu_item_is_selectable (menu_shell->active_menu_item) &&
 	  GTK_MENU_ITEM (menu_shell->active_menu_item)->submenu)
 	{
-	  gtk_menu_shell_select_submenu_first (menu_shell);
+	  if (gtk_menu_shell_select_submenu_first (menu_shell))
+	    break;
 	}
-      else
+
+      /* Try to find a menu running the opposite direction */
+      while (parent_menu_shell && 
+	     (GTK_MENU_SHELL_GET_CLASS (parent_menu_shell)->submenu_placement ==
+	      GTK_MENU_SHELL_GET_CLASS (menu_shell)->submenu_placement))
 	{
-	  /* Try to find a menu running the opposite direction */
-	  while (parent_menu_shell && 
-		 (GTK_MENU_SHELL_GET_CLASS (parent_menu_shell)->submenu_placement ==
-		  GTK_MENU_SHELL_GET_CLASS (menu_shell)->submenu_placement))
-            {
-              GtkWidget *tmp_widget = parent_menu_shell->parent_menu_shell;
+	  GtkWidget *tmp_widget = parent_menu_shell->parent_menu_shell;
+	  
+	  if (tmp_widget)
+	    parent_menu_shell = GTK_MENU_SHELL (tmp_widget);
+	  else
+	    parent_menu_shell = NULL;
+	}
+      
+      if (parent_menu_shell)
+	{
+	  if (PACK_DIRECTION (parent_menu_shell) == GTK_PACK_DIRECTION_LTR)
+	    gtk_menu_shell_move_selected (parent_menu_shell, 1);
+	  else
+	    gtk_menu_shell_move_selected (parent_menu_shell, -1);
 
-              if (tmp_widget)
-                parent_menu_shell = GTK_MENU_SHELL (tmp_widget);
-              else
-                parent_menu_shell = NULL;
-            }
-
-	  if (parent_menu_shell)
-	    {
-	      gtk_menu_shell_move_selected (parent_menu_shell, 1);
-	      gtk_menu_shell_select_submenu_first (parent_menu_shell);
-	    }
+	  gtk_menu_shell_select_submenu_first (parent_menu_shell);
 	}
       break;
       
@@ -1139,7 +1265,6 @@ gtk_real_menu_shell_move_current (GtkMenuShell      *menu_shell,
 	gtk_menu_shell_select_first (menu_shell, TRUE);
       break;
     }
-  
 }
 
 static void
@@ -1147,13 +1272,16 @@ gtk_real_menu_shell_activate_current (GtkMenuShell      *menu_shell,
 				      gboolean           force_hide)
 {
   if (menu_shell->active_menu_item &&
-      _gtk_menu_item_is_selectable (menu_shell->active_menu_item) &&
-      GTK_MENU_ITEM (menu_shell->active_menu_item)->submenu == NULL)
-    {
+      _gtk_menu_item_is_selectable (menu_shell->active_menu_item))
+  {
+   
+    if (GTK_MENU_ITEM (menu_shell->active_menu_item)->submenu == NULL)
       gtk_menu_shell_activate_item (menu_shell,
 				    menu_shell->active_menu_item,
 				    force_hide);
-    }
+    else
+      _gtk_menu_item_popup_submenu (menu_shell->active_menu_item);
+  }
 }
 
 static void
@@ -1220,3 +1348,194 @@ gtk_menu_shell_cancel (GtkMenuShell *menu_shell)
 
   g_signal_emit (menu_shell, menu_shell_signals[CANCEL], 0);
 }
+
+static GtkMnemonicHash *
+gtk_menu_shell_get_mnemonic_hash (GtkMenuShell *menu_shell,
+				  gboolean      create)
+{
+  GtkMenuShellPrivate *private = GTK_MENU_SHELL_GET_PRIVATE (menu_shell);
+
+  if (!private->mnemonic_hash && create)
+    private->mnemonic_hash = _gtk_mnemonic_hash_new ();
+  
+  return private->mnemonic_hash;
+}
+
+static void
+menu_shell_add_mnemonic_foreach (guint    keyval,
+				 GSList  *targets,
+				 gpointer data)
+{
+  GtkKeyHash *key_hash = data;
+
+  _gtk_key_hash_add_entry (key_hash, keyval, 0, GUINT_TO_POINTER (keyval));
+}
+
+static GtkKeyHash *
+gtk_menu_shell_get_key_hash (GtkMenuShell *menu_shell,
+			     gboolean      create)
+{
+  GtkMenuShellPrivate *private = GTK_MENU_SHELL_GET_PRIVATE (menu_shell);
+  GtkWidget *widget = GTK_WIDGET (menu_shell);
+
+  if (!private->key_hash && create && gtk_widget_has_screen (widget))
+    {
+      GtkMnemonicHash *mnemonic_hash = gtk_menu_shell_get_mnemonic_hash (menu_shell, FALSE);
+      GdkScreen *screen = gtk_widget_get_screen (widget);
+      GdkKeymap *keymap = gdk_keymap_get_for_display (gdk_screen_get_display (screen));
+
+      if (!mnemonic_hash)
+	return NULL;
+      
+      private->key_hash = _gtk_key_hash_new (keymap, NULL);
+
+      _gtk_mnemonic_hash_foreach (mnemonic_hash,
+				  menu_shell_add_mnemonic_foreach,
+				  private->key_hash);
+    }
+  
+  return private->key_hash;
+}
+
+static void
+gtk_menu_shell_reset_key_hash (GtkMenuShell *menu_shell)
+{
+  GtkMenuShellPrivate *private = GTK_MENU_SHELL_GET_PRIVATE (menu_shell);
+
+  if (private->key_hash)
+    {
+      _gtk_key_hash_free (private->key_hash);
+      private->key_hash = NULL;
+    }
+}
+
+static gboolean
+gtk_menu_shell_activate_mnemonic (GtkMenuShell *menu_shell,
+				  GdkEventKey  *event)
+{
+  GtkMnemonicHash *mnemonic_hash;
+  GtkKeyHash *key_hash;
+  GSList *entries;
+  gboolean result = FALSE;
+
+  mnemonic_hash = gtk_menu_shell_get_mnemonic_hash (menu_shell, FALSE);
+  if (!mnemonic_hash)
+    return FALSE;
+
+  key_hash = gtk_menu_shell_get_key_hash (menu_shell, TRUE);
+  if (!key_hash)
+    return FALSE;
+  
+  entries = _gtk_key_hash_lookup (key_hash,
+				  event->hardware_keycode,
+				  event->state,
+				  gtk_accelerator_get_default_mod_mask (),
+				  event->group);
+
+  if (entries)
+    result = _gtk_mnemonic_hash_activate (mnemonic_hash,
+					  GPOINTER_TO_UINT (entries->data));
+
+  return result;
+}
+
+void
+_gtk_menu_shell_add_mnemonic (GtkMenuShell *menu_shell,
+			      guint      keyval,
+			      GtkWidget *target)
+{
+  g_return_if_fail (GTK_IS_MENU_SHELL (menu_shell));
+  g_return_if_fail (GTK_IS_WIDGET (target));
+
+  _gtk_mnemonic_hash_add (gtk_menu_shell_get_mnemonic_hash (menu_shell, TRUE),
+			  keyval, target);
+  gtk_menu_shell_reset_key_hash (menu_shell);
+}
+
+void
+_gtk_menu_shell_remove_mnemonic (GtkMenuShell *menu_shell,
+				 guint      keyval,
+				 GtkWidget *target)
+{
+  g_return_if_fail (GTK_IS_MENU_SHELL (menu_shell));
+  g_return_if_fail (GTK_IS_WIDGET (target));
+  
+  _gtk_mnemonic_hash_remove (gtk_menu_shell_get_mnemonic_hash (menu_shell, TRUE),
+			     keyval, target);
+  gtk_menu_shell_reset_key_hash (menu_shell);
+}
+
+/**
+ * gtk_menu_shell_get_take_focus:
+ * @menu_shell: a #GtkMenuShell
+ *
+ * Returns %TRUE if the menu shell will take the keyboard focus on popup.
+ *
+ * Returns: %TRUE if the menu shell will take the keyboard focus on popup.
+ *
+ * Since: 2.8
+ **/
+gboolean
+gtk_menu_shell_get_take_focus (GtkMenuShell *menu_shell)
+{
+  GtkMenuShellPrivate *priv;
+
+  g_return_val_if_fail (GTK_IS_MENU_SHELL (menu_shell), FALSE);
+
+  priv = GTK_MENU_SHELL_GET_PRIVATE (menu_shell);
+
+  return priv->take_focus;
+}
+
+/**
+ * gtk_menu_shell_set_take_focus:
+ * @menu_shell: a #GtkMenuShell
+ * @take_focus: %TRUE if the menu shell should take the keyboard focus on popup.
+ *
+ * If @take_focus is %TRUE (the default) the menu shell will take the keyboard 
+ * focus so that it will receive all keyboard events which is needed to enable
+ * keyboard navigation in menus.
+ *
+ * Setting @take_focus to %FALSE is useful only for special applications
+ * like virtual keyboard implementations which should not take keyboard
+ * focus.
+ *
+ * The @take_focus state of a menu or menu bar is automatically propagated
+ * to submenus whenever a submenu is popped up, so you don't have to worry
+ * about recursively setting it for your entire menu hierarchy. Only when
+ * programmatically picking a submenu and popping it up manually, the
+ * @take_focus property of the submenu needs to be set explicitely.
+ *
+ * Note that setting it to %FALSE has side-effects:
+ *
+ * If the focus is in some other app, it keeps the focus and keynav in
+ * the menu doesn't work. Consequently, keynav on the menu will only
+ * work if the focus is on some toplevel owned by the onscreen keyboard.
+ *
+ * To avoid confusing the user, menus with @take_focus set to %FALSE
+ * should not display mnemonics or accelerators, since it cannot be
+ * guaranteed that they will work.
+ *
+ * See also gdk_keyboard_grab()
+ *
+ * Since: 2.8
+ **/
+void
+gtk_menu_shell_set_take_focus (GtkMenuShell *menu_shell,
+                               gboolean      take_focus)
+{
+  GtkMenuShellPrivate *priv;
+
+  g_return_if_fail (GTK_IS_MENU_SHELL (menu_shell));
+
+  priv = GTK_MENU_SHELL_GET_PRIVATE (menu_shell);
+
+  if (priv->take_focus != take_focus)
+    {
+      priv->take_focus = take_focus;
+      g_object_notify (G_OBJECT (menu_shell), "take-focus");
+    }
+}
+
+#define __GTK_MENU_SHELL_C__
+#include "gtkaliasdef.c"
