@@ -19,12 +19,13 @@
  */
 
 /*
- * Modified by the GTK+ Team and others 1997-2000.  See the AUTHORS
+ * Modified by the GTK+ Team and others 1997-2004.  See the AUTHORS
  * file for a list of people on the GTK+ Team.  See the ChangeLog
  * files for a list of changes.  These files are distributed with
  * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
  */
 
+#include <config.h>
 #include <stdio.h>
 #include <math.h>
 #include "gtkintl.h"
@@ -33,6 +34,8 @@
 #include "gtkrange.h"
 #include "gtkintl.h"
 #include "gtkscrollbar.h"
+#include "gtkprivate.h"
+#include "gtkalias.h"
 
 #define SCROLL_INITIAL_DELAY 250  /* must hold button this long before ... */
 #define SCROLL_LATER_DELAY   100  /* ... it starts repeating at this rate  */
@@ -49,6 +52,7 @@ enum {
   VALUE_CHANGED,
   ADJUST_BOUNDS,
   MOVE_SLIDER,
+  CHANGE_VALUE,
   LAST_SIGNAL
 };
 
@@ -131,7 +135,7 @@ static void gtk_range_style_set      (GtkWidget        *widget,
 static void update_slider_position   (GtkRange	       *range,
 				      gint              mouse_x,
 				      gint              mouse_y);
-
+static void stop_scrolling           (GtkRange         *range);
 
 /* Range methods */
 
@@ -171,7 +175,8 @@ static void          gtk_range_reset_update_timer       (GtkRange      *range);
 static void          gtk_range_remove_update_timer      (GtkRange      *range);
 static GdkRectangle* get_area                           (GtkRange      *range,
                                                          MouseLocation  location);
-static void          gtk_range_internal_set_value       (GtkRange      *range,
+static gboolean      gtk_range_real_change_value        (GtkRange      *range,
+                                                         GtkScrollType  scroll,
                                                          gdouble        value);
 static void          gtk_range_update_value             (GtkRange      *range);
 
@@ -244,6 +249,7 @@ gtk_range_class_init (GtkRangeClass *class)
   widget_class->style_set = gtk_range_style_set;
 
   class->move_slider = gtk_range_move_slider;
+  class->change_value = gtk_range_real_change_value;
 
   class->slider_detail = "slider";
   class->stepper_detail = "stepper";
@@ -276,15 +282,52 @@ gtk_range_class_init (GtkRangeClass *class)
                   _gtk_marshal_VOID__ENUM,
                   G_TYPE_NONE, 1,
                   GTK_TYPE_SCROLL_TYPE);
+
+  /**
+   * GtkRange::change-value:
+   * @range: the range that received the signal.
+   * @scroll: the type of scroll action that was performed.
+   * @value: the new value resulting from the scroll action.
+   * @returns: %TRUE to prevent other handlers from being invoked for the
+   * signal.  %FALSE to propagate the signal further.
+   *
+   * The ::change-value signal is emitted when a scroll action is
+   * performed on a range.  It allows an application to determine the
+   * type of scroll event that occurred and the resultant new value.
+   * The application can handle the event itself and return %TRUE to
+   * prevent further processing.  Or, by returning %FALSE, it can pass
+   * the event to other handlers until the default GTK+ handler is
+   * reached.
+   *
+   * The value parameter is unrounded.  An application that overrides
+   * the ::change-value signal is responsible for clamping the value to
+   * the desired number of decimal digits; the default GTK+ handler 
+   * clamps the value based on @range->round_digits.
+   *
+   * It is not possible to use delayed update policies in an overridden
+   * ::change-value handler.
+   *
+   * Since: 2.6
+   */
+  signals[CHANGE_VALUE] =
+    g_signal_new ("change_value",
+                  G_TYPE_FROM_CLASS (gobject_class),
+                  G_SIGNAL_RUN_LAST,
+                  G_STRUCT_OFFSET (GtkRangeClass, change_value),
+                  _gtk_boolean_handled_accumulator, NULL,
+                  _gtk_marshal_BOOLEAN__ENUM_DOUBLE,
+                  G_TYPE_BOOLEAN, 2,
+                  GTK_TYPE_SCROLL_TYPE,
+                  G_TYPE_DOUBLE);
   
   g_object_class_install_property (gobject_class,
                                    PROP_UPDATE_POLICY,
-                                   g_param_spec_enum ("update_policy",
+                                   g_param_spec_enum ("update-policy",
 						      P_("Update policy"),
 						      P_("How the range should be updated on the screen"),
 						      GTK_TYPE_UPDATE_TYPE,
 						      GTK_UPDATE_CONTINUOUS,
-						      G_PARAM_READWRITE));
+						      GTK_PARAM_READWRITE));
   
   g_object_class_install_property (gobject_class,
                                    PROP_ADJUSTMENT,
@@ -292,7 +335,7 @@ gtk_range_class_init (GtkRangeClass *class)
 							P_("Adjustment"),
 							P_("The GtkAdjustment that contains the current value of this range object"),
                                                         GTK_TYPE_ADJUSTMENT,
-                                                        G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
+                                                        GTK_PARAM_READWRITE | G_PARAM_CONSTRUCT));
 
   g_object_class_install_property (gobject_class,
                                    PROP_INVERTED,
@@ -300,56 +343,56 @@ gtk_range_class_init (GtkRangeClass *class)
 							P_("Inverted"),
 							P_("Invert direction slider moves to increase range value"),
                                                          FALSE,
-                                                         G_PARAM_READWRITE));
+                                                         GTK_PARAM_READWRITE));
   
   gtk_widget_class_install_style_property (widget_class,
-					   g_param_spec_int ("slider_width",
+					   g_param_spec_int ("slider-width",
 							     P_("Slider Width"),
 							     P_("Width of scrollbar or scale thumb"),
 							     0,
 							     G_MAXINT,
 							     14,
-							     G_PARAM_READABLE));
+							     GTK_PARAM_READABLE));
   gtk_widget_class_install_style_property (widget_class,
-					   g_param_spec_int ("trough_border",
+					   g_param_spec_int ("trough-border",
                                                              P_("Trough Border"),
                                                              P_("Spacing between thumb/steppers and outer trough bevel"),
                                                              0,
                                                              G_MAXINT,
                                                              1,
-                                                             G_PARAM_READABLE));
+                                                             GTK_PARAM_READABLE));
   gtk_widget_class_install_style_property (widget_class,
-					   g_param_spec_int ("stepper_size",
+					   g_param_spec_int ("stepper-size",
 							     P_("Stepper Size"),
 							     P_("Length of step buttons at ends"),
 							     0,
 							     G_MAXINT,
 							     14,
-							     G_PARAM_READABLE));
+							     GTK_PARAM_READABLE));
   gtk_widget_class_install_style_property (widget_class,
-					   g_param_spec_int ("stepper_spacing",
+					   g_param_spec_int ("stepper-spacing",
 							     P_("Stepper Spacing"),
 							     P_("Spacing between step buttons and thumb"),
                                                              0,
 							     G_MAXINT,
 							     0,
-							     G_PARAM_READABLE));
+							     GTK_PARAM_READABLE));
   gtk_widget_class_install_style_property (widget_class,
-					   g_param_spec_int ("arrow_displacement_x",
+					   g_param_spec_int ("arrow-displacement-x",
 							     P_("Arrow X Displacement"),
 							     P_("How far in the x direction to move the arrow when the button is depressed"),
 							     G_MININT,
 							     G_MAXINT,
 							     0,
-							     G_PARAM_READABLE));
+							     GTK_PARAM_READABLE));
   gtk_widget_class_install_style_property (widget_class,
-					   g_param_spec_int ("arrow_displacement_y",
+					   g_param_spec_int ("arrow-displacement-y",
 							     P_("Arrow Y Displacement"),
 							     P_("How far in the y direction to move the arrow when the button is depressed"),
 							     G_MININT,
 							     G_MAXINT,
 							     0,
-							     G_PARAM_READABLE));
+							     GTK_PARAM_READABLE));
 }
 
 static void
@@ -477,7 +520,7 @@ gtk_range_set_update_policy (GtkRange      *range,
   if (range->update_policy != policy)
     {
       range->update_policy = policy;
-      g_object_notify (G_OBJECT (range), "update_policy");
+      g_object_notify (G_OBJECT (range), "update-policy");
     }
 }
 
@@ -838,8 +881,6 @@ gtk_range_map (GtkWidget *widget)
 {
   GtkRange *range = GTK_RANGE (widget);
   
-  g_return_if_fail (GTK_IS_RANGE (widget));
-
   gdk_window_show (range->event_window);
 
   GTK_WIDGET_CLASS (parent_class)->map (widget);
@@ -850,7 +891,7 @@ gtk_range_unmap (GtkWidget *widget)
 {
   GtkRange *range = GTK_RANGE (widget);
     
-  g_return_if_fail (GTK_IS_RANGE (widget));
+  stop_scrolling (range);
 
   gdk_window_hide (range->event_window);
 
@@ -1308,7 +1349,8 @@ update_slider_position (GtkRange *range,
   gint delta;
   gint c;
   gdouble new_value;
-  
+  gboolean handled;
+
   if (range->orientation == GTK_ORIENTATION_VERTICAL)
     delta = mouse_y - range->slide_initial_coordinate;
   else
@@ -1318,10 +1360,12 @@ update_slider_position (GtkRange *range,
 
   new_value = coord_to_value (range, c);
   
-  gtk_range_internal_set_value (range, new_value);
+  g_signal_emit (range, signals[CHANGE_VALUE], 0, GTK_SCROLL_JUMP, new_value,
+                 &handled);
 }
 
-static void stop_scrolling (GtkRange *range)
+static void 
+stop_scrolling (GtkRange *range)
 {
   range_grab_remove (range);
   gtk_range_remove_step_timer (range);
@@ -1409,9 +1453,13 @@ gtk_range_scroll_event (GtkWidget      *widget,
     {
       GtkAdjustment *adj = GTK_RANGE (range)->adjustment;
       gdouble delta;
+      gboolean handled;
 
       delta = _gtk_range_get_wheel_delta (range, event->direction);
-      gtk_range_internal_set_value (range, adj->value + delta);
+
+      g_signal_emit (range, signals[CHANGE_VALUE], 0,
+                     GTK_SCROLL_JUMP, adj->value + delta,
+                     &handled);
       
       /* Policy DELAYED makes sense with scroll events,
        * but DISCONTINUOUS doesn't, so we update immediately
@@ -1551,18 +1599,22 @@ static void
 step_back (GtkRange *range)
 {
   gdouble newval;
+  gboolean handled;
   
   newval = range->adjustment->value - range->adjustment->step_increment;
-  gtk_range_internal_set_value (range, newval);
+  g_signal_emit (range, signals[CHANGE_VALUE], 0,
+                 GTK_SCROLL_STEP_BACKWARD, newval, &handled);
 }
 
 static void
 step_forward (GtkRange *range)
 {
   gdouble newval;
+  gboolean handled;
 
   newval = range->adjustment->value + range->adjustment->step_increment;
-  gtk_range_internal_set_value (range, newval);
+  g_signal_emit (range, signals[CHANGE_VALUE], 0,
+                 GTK_SCROLL_STEP_FORWARD, newval, &handled);
 }
 
 
@@ -1570,33 +1622,42 @@ static void
 page_back (GtkRange *range)
 {
   gdouble newval;
+  gboolean handled;
 
   newval = range->adjustment->value - range->adjustment->page_increment;
-  gtk_range_internal_set_value (range, newval);
+  g_signal_emit (range, signals[CHANGE_VALUE], 0,
+                 GTK_SCROLL_PAGE_BACKWARD, newval, &handled);
 }
 
 static void
 page_forward (GtkRange *range)
 {
   gdouble newval;
+  gboolean handled;
 
   newval = range->adjustment->value + range->adjustment->page_increment;
-  gtk_range_internal_set_value (range, newval);
+  g_signal_emit (range, signals[CHANGE_VALUE], 0,
+                 GTK_SCROLL_PAGE_FORWARD, newval, &handled);
 }
 
 static void
 scroll_begin (GtkRange *range)
 {
-  gtk_range_internal_set_value (range, range->adjustment->lower);
+  gboolean handled;
+  g_signal_emit (range, signals[CHANGE_VALUE], 0,
+                 GTK_SCROLL_START, range->adjustment->lower,
+                 &handled);
 }
 
 static void
 scroll_end (GtkRange *range)
 {
   gdouble newval;
+  gboolean handled;
 
   newval = range->adjustment->upper - range->adjustment->page_size;
-  gtk_range_internal_set_value (range, newval);
+  g_signal_emit (range, signals[CHANGE_VALUE], 0, GTK_SCROLL_END, newval,
+                 &handled);
 }
 
 static void
@@ -1722,12 +1783,12 @@ gtk_range_get_props (GtkRange  *range,
   gint tmp_arrow_displacement_x, tmp_arrow_displacement_y;
   
   gtk_widget_style_get (widget,
-                        "slider_width", &tmp_slider_width,
-                        "trough_border", &tmp_trough_border,
-                        "stepper_size", &tmp_stepper_size,
-                        "stepper_spacing", &tmp_stepper_spacing,
-			"arrow_displacement_x", &tmp_arrow_displacement_x,
-			"arrow_displacement_y", &tmp_arrow_displacement_y,
+                        "slider-width", &tmp_slider_width,
+                        "trough-border", &tmp_trough_border,
+                        "stepper-size", &tmp_stepper_size,
+                        "stepper-spacing", &tmp_stepper_spacing,
+			"arrow-displacement-x", &tmp_arrow_displacement_x,
+			"arrow-displacement-y", &tmp_arrow_displacement_y,
                         NULL);
   
   if (GTK_WIDGET_CAN_FOCUS (range))
@@ -2320,9 +2381,10 @@ get_area (GtkRange     *range,
   return NULL;
 }
 
-static void
-gtk_range_internal_set_value (GtkRange *range,
-                              gdouble   value)
+static gboolean
+gtk_range_real_change_value (GtkRange     *range,
+                             GtkScrollType scroll,
+                             gdouble       value)
 {
   /* potentially adjust the bounds _before we clamp */
   g_signal_emit (range, signals[ADJUST_BOUNDS], 0, value);
@@ -2332,12 +2394,15 @@ gtk_range_internal_set_value (GtkRange *range,
 
   if (range->round_digits >= 0)
     {
-      char buffer[128];
+      gdouble power;
+      gint i;
 
-      /* This is just so darn lame. */
-      g_snprintf (buffer, 128, "%0.*f",
-                  range->round_digits, value);
-      sscanf (buffer, "%lf", &value);
+      i = range->round_digits;
+      power = 1;
+      while (i--)
+        power *= 10;
+      
+      value = floor ((value * power) + 0.5) / power;
     }
   
   if (range->adjustment->value != value)
@@ -2365,6 +2430,7 @@ gtk_range_internal_set_value (GtkRange *range,
           break;
         }
     }
+  return FALSE;
 }
 
 static void
@@ -2482,3 +2548,6 @@ gtk_range_remove_update_timer (GtkRange *range)
       range->update_timeout_id = 0;
     }
 }
+
+#define __GTK_RANGE_C__
+#include "gtkaliasdef.c"
