@@ -23,6 +23,7 @@
 #include "gtkintl.h"
 #include "gtktreednd.h"
 #include "gtkalias.h"
+#include "gtkprivate.h"
 #include <string.h>
 
 /* ITER FORMAT:
@@ -323,19 +324,19 @@ gtk_tree_model_filter_class_init (GtkTreeModelFilterClass *filter_class)
    */
   g_object_class_install_property (object_class,
                                    PROP_CHILD_MODEL,
-                                   g_param_spec_object ("child_model",
+                                   g_param_spec_object ("child-model",
                                                         ("The child model"),
                                                         ("The model for the filtermodel to filter"),
                                                         GTK_TYPE_TREE_MODEL,
-                                                        G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+                                                        GTK_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 
   g_object_class_install_property (object_class,
                                    PROP_VIRTUAL_ROOT,
-                                   g_param_spec_boxed ("virtual_root",
+                                   g_param_spec_boxed ("virtual-root",
                                                        ("The virtual root"),
                                                        ("The virtual root (relative to the child model) for this filtermodel"),
                                                        GTK_TYPE_TREE_PATH,
-                                                       G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+                                                       GTK_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 
   g_type_class_add_private (object_class, sizeof (GtkTreeModelFilterPrivate));
 }
@@ -826,9 +827,7 @@ gtk_tree_model_filter_fetch_child (GtkTreeModelFilter *filter,
   g_array_insert_val (level->array, i, elt);
   *index = i;
 
-  if (i > 0)
-    i--;
-  for ( ; i < level->array->len; i++)
+  for (i = MAX (--i, 0); i < level->array->len; i++)
     {
       FilterElt *e = &(g_array_index (level->array, FilterElt, i));
       if (e->children)
@@ -1144,7 +1143,7 @@ gtk_tree_model_filter_row_changed (GtkTreeModel *c_model,
 
   if (!path)
     /* parent is probably being filtered out */
-    goto done;    
+    goto done;
 
   gtk_tree_model_get_iter (GTK_TREE_MODEL (filter), &iter, path);
 
@@ -2423,8 +2422,8 @@ gtk_tree_model_filter_new (GtkTreeModel *child_model,
   g_return_val_if_fail (GTK_IS_TREE_MODEL (child_model), NULL);
 
   retval = g_object_new (GTK_TYPE_TREE_MODEL_FILTER, 
-			 "child_model", child_model,
-			 "virtual_root", root,
+			 "child-model", child_model,
+			 "virtual-root", root,
 			 NULL);
 
   return retval;
@@ -2869,7 +2868,7 @@ gtk_tree_model_filter_refilter (GtkTreeModelFilter *filter)
  * gtk_tree_model_ref_node(). This might be useful if the child model
  * being filtered is static (and doesn't change often) and there has been
  * a lot of unreffed access to nodes. As a side effect of this function,
- * all unreffed iters will be invalid.
+ * all unreffed itters will be invalid.
  *
  * Since: 2.4
  */
