@@ -42,11 +42,11 @@ typedef enum
 } GtkTreeViewDropPosition;
 
 #define GTK_TYPE_TREE_VIEW		(gtk_tree_view_get_type ())
-#define GTK_TREE_VIEW(obj)		(GTK_CHECK_CAST ((obj), GTK_TYPE_TREE_VIEW, GtkTreeView))
-#define GTK_TREE_VIEW_CLASS(klass)	(GTK_CHECK_CLASS_CAST ((klass), GTK_TYPE_TREE_VIEW, GtkTreeViewClass))
-#define GTK_IS_TREE_VIEW(obj)		(GTK_CHECK_TYPE ((obj), GTK_TYPE_TREE_VIEW))
-#define GTK_IS_TREE_VIEW_CLASS(klass)	(GTK_CHECK_CLASS_TYPE ((obj), GTK_TYPE_TREE_VIEW))
-#define GTK_TREE_VIEW_GET_CLASS(obj)    (GTK_CHECK_GET_CLASS ((obj), GTK_TYPE_TREE_VIEW, GtkTreeViewClass))
+#define GTK_TREE_VIEW(obj)		(G_TYPE_CHECK_INSTANCE_CAST ((obj), GTK_TYPE_TREE_VIEW, GtkTreeView))
+#define GTK_TREE_VIEW_CLASS(klass)	(G_TYPE_CHECK_CLASS_CAST ((klass), GTK_TYPE_TREE_VIEW, GtkTreeViewClass))
+#define GTK_IS_TREE_VIEW(obj)		(G_TYPE_CHECK_INSTANCE_TYPE ((obj), GTK_TYPE_TREE_VIEW))
+#define GTK_IS_TREE_VIEW_CLASS(klass)	(G_TYPE_CHECK_CLASS_TYPE ((klass), GTK_TYPE_TREE_VIEW))
+#define GTK_TREE_VIEW_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS ((obj), GTK_TYPE_TREE_VIEW, GtkTreeViewClass))
 
 typedef struct _GtkTreeView           GtkTreeView;
 typedef struct _GtkTreeViewClass      GtkTreeViewClass;
@@ -87,19 +87,27 @@ struct _GtkTreeViewClass
   void     (* cursor_changed)             (GtkTreeView       *tree_view);
 
   /* Key Binding signals */
-  void     (* move_cursor)                (GtkTreeView       *tree_view,
+  gboolean (* move_cursor)                (GtkTreeView       *tree_view,
 				           GtkMovementStep    step,
 				           gint               count);
-  void     (* select_all)                 (GtkTreeView       *tree_view);
-  void     (* select_cursor_row)          (GtkTreeView       *tree_view,
+  gboolean (* select_all)                 (GtkTreeView       *tree_view);
+  gboolean (* unselect_all)               (GtkTreeView       *tree_view);
+  gboolean (* select_cursor_row)          (GtkTreeView       *tree_view,
 					   gboolean           start_editing);
-  void     (* toggle_cursor_row)          (GtkTreeView       *tree_view);
-  void     (* expand_collapse_cursor_row) (GtkTreeView       *tree_view,
+  gboolean (* toggle_cursor_row)          (GtkTreeView       *tree_view);
+  gboolean (* expand_collapse_cursor_row) (GtkTreeView       *tree_view,
 					   gboolean           logical,
 					   gboolean           expand,
 					   gboolean           open_all);
-  void     (* select_cursor_parent)       (GtkTreeView       *tree_view);
-  void     (* start_interactive_search)   (GtkTreeView       *tree_view); 
+  gboolean (* select_cursor_parent)       (GtkTreeView       *tree_view);
+  gboolean (* start_interactive_search)   (GtkTreeView       *tree_view);
+
+  /* Padding for future expansion */
+  void (*_gtk_reserved0) (void);
+  void (*_gtk_reserved1) (void);
+  void (*_gtk_reserved2) (void);
+  void (*_gtk_reserved3) (void);
+  void (*_gtk_reserved4) (void);
 };
 
 
@@ -116,10 +124,13 @@ typedef gboolean (*GtkTreeViewSearchEqualFunc) (GtkTreeModel            *model,
 						const gchar             *key,
 						GtkTreeIter             *iter,
 						gpointer                 search_data);
+typedef gboolean (*GtkTreeViewRowSeparatorFunc) (GtkTreeModel      *model,
+						 GtkTreeIter       *iter,
+						 gpointer           data);
 
 
 /* Creators */
-GtkType                gtk_tree_view_get_type                      (void);
+GType                  gtk_tree_view_get_type                      (void) G_GNUC_CONST;
 GtkWidget             *gtk_tree_view_new                           (void);
 GtkWidget             *gtk_tree_view_new_with_model                (GtkTreeModel              *model);
 
@@ -154,12 +165,12 @@ gint                   gtk_tree_view_insert_column                 (GtkTreeView
 								    gint                       position);
 gint                   gtk_tree_view_insert_column_with_attributes (GtkTreeView               *tree_view,
 								    gint                       position,
-								    gchar                     *title,
+								    const gchar               *title,
 								    GtkCellRenderer           *cell,
-								    ...);
+								    ...) G_GNUC_NULL_TERMINATED;
 gint                   gtk_tree_view_insert_column_with_data_func  (GtkTreeView               *tree_view,
 								    gint                       position,
-								    gchar                     *title,
+								    const gchar               *title,
 								    GtkCellRenderer           *cell,
                                                                     GtkTreeCellDataFunc        func,
                                                                     gpointer                   data,
@@ -193,6 +204,8 @@ void                   gtk_tree_view_row_activated                 (GtkTreeView
 								    GtkTreeViewColumn         *column);
 void                   gtk_tree_view_expand_all                    (GtkTreeView               *tree_view);
 void                   gtk_tree_view_collapse_all                  (GtkTreeView               *tree_view);
+void                   gtk_tree_view_expand_to_path                (GtkTreeView               *tree_view,
+								    GtkTreePath               *path);
 gboolean               gtk_tree_view_expand_row                    (GtkTreeView               *tree_view,
 								    GtkTreePath               *path,
 								    gboolean                   open_all);
@@ -210,6 +223,11 @@ void                   gtk_tree_view_set_cursor                    (GtkTreeView
 								    GtkTreePath               *path,
 								    GtkTreeViewColumn         *focus_column,
 								    gboolean                   start_editing);
+void                   gtk_tree_view_set_cursor_on_cell            (GtkTreeView               *tree_view,
+								    GtkTreePath               *path,
+								    GtkTreeViewColumn         *focus_column,
+								    GtkCellRenderer           *focus_cell,
+								    gboolean                   start_editing);
 void                   gtk_tree_view_get_cursor                    (GtkTreeView               *tree_view,
 								    GtkTreePath              **path,
 								    GtkTreeViewColumn        **focus_column);
@@ -298,6 +316,21 @@ void gtk_tree_view_set_destroy_count_func (GtkTreeView             *tree_view,
 					   gpointer                 data,
 					   GtkDestroyNotify         destroy);
 
+void     gtk_tree_view_set_fixed_height_mode (GtkTreeView          *tree_view,
+					      gboolean              enable);
+gboolean gtk_tree_view_get_fixed_height_mode (GtkTreeView          *tree_view);
+void     gtk_tree_view_set_hover_selection   (GtkTreeView          *tree_view,
+					      gboolean              hover);
+gboolean gtk_tree_view_get_hover_selection   (GtkTreeView          *tree_view);
+void     gtk_tree_view_set_hover_expand      (GtkTreeView          *tree_view,
+					      gboolean              expand);
+gboolean gtk_tree_view_get_hover_expand      (GtkTreeView          *tree_view);
+
+GtkTreeViewRowSeparatorFunc gtk_tree_view_get_row_separator_func (GtkTreeView               *tree_view);
+void                        gtk_tree_view_set_row_separator_func (GtkTreeView                *tree_view,
+								  GtkTreeViewRowSeparatorFunc func,
+								  gpointer                    data,
+								  GtkDestroyNotify            destroy);
 
 #ifdef __cplusplus
 }
